from __future__ import absolute_import as _
from __future__ import division as _
from __future__ import print_function as _
from __future__ import unicode_literals as _

import numpy as np

from coremltools.models.neural_network import NeuralNetworkBuilder  # type: ignore
from coremltools.models import datatypes, MLModel  # type: ignore
from coremltools.proto import FeatureTypes_pb2 as ft  # type: ignore
from coremltools import (
    _MINIMUM_CUSTOM_LAYER_SPEC_VERSION as IOS_11_2_SPEC_VERSION,
)  # iOS 11.2
from coremltools import (
    _MINIMUM_CUSTOM_MODEL_SPEC_VERSION as IOS_12_SPEC_VERSION,
)  # iOS 12.0
from coremltools import _MINIMUM_NDARRAY_SPEC_VERSION as IOS_13_SPEC_VERSION  # iOS 13.0
from coremltools import __version__ as ct_version
from coremltools.models import _METADATA_VERSION, _METADATA_SOURCE

from coremltools._deps import _HAS_ONNX

# ML model passes
from coremltools.converters.mil.backend.nn.passes.mlmodel_passes import (
    remove_disconnected_layers,
    transform_conv_crop,
)

if _HAS_ONNX:
    import onnx
    from onnx import shape_inference
    from onnx import TensorProto

    from typing import Tuple
    from typing import Text, Union, Optional, Dict, Any, Iterable, Sequence, Callable, List

    from ._operators import (
        _convert_node,
        _SEQUENCE_LAYERS_REGISTRY,
        _ONNX_NODE_REGISTRY,
        _add_const_inputs_if_required,
    )
    from ._operators_nd import _ONNX_NODE_REGISTRY_ND, _convert_node_nd

    from ._graph import Graph, EdgeInfo, Transformer

    from ._transformers import (
        ConvAddFuser,
        DropoutRemover,
        ReshapeInitTensorFuser,
        BNBroadcastedMulFuser,
        BNBroadcastedAddFuser,
        PixelShuffleFuser,
        OutputRenamer,
        AddModelInputsOutputs,
        ConstantsToInitializers,
        ImageScalerRemover,
        ShapeOpRemover,
        ConstantRemover,
        ConstantFillToInitializers,
        ReshapeTransposeReshape_pattern1,
        CastOpRemover,
        DeadCodeElimination,
        PaddingOpRemover,
    )

    from ._error_utils import ErrorHandling
    from ._graph_viz import plot_graph  # type: ignore


USE_SHAPE_MAPPING = True

DEBUG = False


class SupportedVersion:
    # Supported iOS Version
    # New OS Version must be added at the end to maintain backward version index
    supported_ios_version = ["11.2", "12", "13"]
    IOS_13_VERSION = supported_ios_version.index("13")
    ND_ARRARY_SUPPORT = IOS_13_VERSION

    @staticmethod
    def ios_support_check(minimum_ios_deployment_target):
        return minimum_ios_deployment_target in SupportedVersion.supported_ios_version

    @staticmethod
    def is_nd_array_supported(minimum_ios_deployment_target):
        if not SupportedVersion.ios_support_check(minimum_ios_deployment_target):
            raise TypeError(
                "{} not supported. Please provide one of target iOS: {}".format(
                minimum_ios_deployment_target,
                SupportedVersion.supported_ios_version
                )
            )

        minimum_ios_deployment_target_index = SupportedVersion.supported_ios_version.index(
            minimum_ios_deployment_target
        )
        return SupportedVersion.ND_ARRARY_SUPPORT <= minimum_ios_deployment_target_index

    @staticmethod
    def get_supported_ios():
        return SupportedVersion.supported_ios_version

    @staticmethod
    def get_specification_version(minimum_ios_deployment_target):
        if not SupportedVersion.ios_support_check(minimum_ios_deployment_target):
            raise TypeError(
                "{} not supported. Please provide one of target iOS: {}",
                minimum_ios_deployment_target,
                SupportedVersion.supported_ios_version,
            )

        if minimum_ios_deployment_target == "11.2":
            return IOS_11_2_SPEC_VERSION
        elif minimum_ios_deployment_target == "12":
            return IOS_12_SPEC_VERSION
        else:
            return IOS_13_SPEC_VERSION


"""
inputs: list of tuples.
      [Tuple]: [(name, type, shape)]
"""


def _make_coreml_input_features(
    graph, onnx_coreml_input_shape_map, disable_coreml_rank5_mapping=False
):  # type: (...) -> Sequence[Tuple[Text, datatypes.Array]]
    """
    If "disable_coreml_rank5_mapping" is False, then:

    ONNX shapes to CoreML static shapes mapping
    length==1: [C]
    length==2: [B,C]
    length==3: [C,H,W] or [Seq,B,C]
    length==4: [B,C,H,W]

    If "disable_coreml_rank5_mapping" is True, then
    onnx shapes are mapped "as is" to CoreML.
    """
    inputs = graph.inputs
    op_types = graph.blob_to_op_type
    features = []
    for input_ in inputs:
        shape = input_[2]
        if disable_coreml_rank5_mapping:
            if len(shape) > 5:
                raise ValueError(
                    "ONNX input %s has a rank greater than 5, which is not supported in CoreML framework"
                    % str(input_[0])
                )
            else:
                features.append((str(input_[0]), datatypes.Array(*shape)))
            continue

        if USE_SHAPE_MAPPING and input_[0] in onnx_coreml_input_shape_map:
            mapp = onnx_coreml_input_shape_map[input_[0]]
            if len(mapp) != len(shape):
                raise ValueError(
                    "Incorrect value in onnx_coreml_input_shape_map argument"
                )
            graph.onnx_coreml_shape_mapping[input_[0]] = mapp
            coreml_shape = [1, 1, 1]
            for i in range(3):
                if (i + 2) in mapp:
                    coreml_shape[i] = shape[mapp.index(i + 2)]
            shape = coreml_shape
        else:
            if len(shape) == 0:
                shape = [1, 1, 1]
            elif len(shape) == 1:
                # assume [C]
                if USE_SHAPE_MAPPING:
                    graph.onnx_coreml_shape_mapping[input_[0]] = [2]
            elif len(shape) == 2:
                # assume [Batch,C]
                shape = [shape[1]]
                if USE_SHAPE_MAPPING:
                    graph.onnx_coreml_shape_mapping[input_[0]] = [1, 2]
            elif len(shape) == 3:
                # assume [C,H,W] unless its connected an op that bestows another mapping
                if input_[0] in op_types and len(op_types[input_[0]]) == 1:
                    if str(op_types[input_[0]][0]) in _SEQUENCE_LAYERS_REGISTRY:
                        # (Seq,B,C)
                        shape = [shape[2]]
                        if USE_SHAPE_MAPPING:
                            graph.onnx_coreml_shape_mapping[input_[0]] = [0, 1, 2]
                    elif str(op_types[input_[0]][0]) in [
                        "MaxPool",
                        "AveragePool",
                        "BatchNormalization",
                        "GlobalAveragePool",
                        "GlobalLpPool",
                        "GlobalMaxPool",
                        "InstanceNormalization",
                        "LRN",
                        "LpPool",
                        "Conv",
                        "ConvTranspose",
                    ]:
                        # (B,C,W)
                        shape = [shape[1], 1, shape[2]]
                        if USE_SHAPE_MAPPING:
                            graph.onnx_coreml_shape_mapping[input_[0]] = [1, 2, 4]
                    else:
                        if USE_SHAPE_MAPPING:
                            graph.onnx_coreml_shape_mapping[input_[0]] = [2, 3, 4]
                else:
                    if USE_SHAPE_MAPPING:
                        graph.onnx_coreml_shape_mapping[input_[0]] = [2, 3, 4]
            elif len(shape) == 4:  # (B,C,H,W) --> (C,H,W)
                shape = shape[1:]
                if USE_SHAPE_MAPPING:
                    graph.onnx_coreml_shape_mapping[input_[0]] = [1, 2, 3, 4]
            else:
                raise ValueError(
                    "CoreML input cannot be more than rank 4. Input shape: %s, input: '%s' "
                    % (str(shape), str(input_[0]))
                )
        features.append((str(input_[0]), datatypes.Array(*shape)))
    return features


"""
outputs: list of tuples.
      [Tuple]: [(name, type, shape)]
"""


def _make_coreml_output_features(
    graph, forceShape=False, disable_coreml_rank5_mapping=False
):  # type: (...) -> Sequence[Tuple[Text, datatypes.Array]]
    features = []
    outputs = graph.outputs
    op_types = graph.blob_from_op_type
    ops_allowing_zerod_output = {"Size"}

    for output_ in outputs:
        if op_types[output_[0]] in ops_allowing_zerod_output and len(output_[2]) == 0:
            output_ = list(output_)
            output_[2] = (1,)

        if disable_coreml_rank5_mapping:
            shape = output_[2]
            if len(shape) > 5:
                raise ValueError(
                    "ONNX output %s has a rank greater than 5, which is not supported in CoreML framework"
                    % str(output_[0])
                )
            else:
                features.append((str(output_[0]), datatypes.Array(*shape)))
            continue

        if not forceShape:
            features.append((str(output_[0]), None))
        else:
            shape = output_[2]
            if len(shape) == 0:
                shape = [1, 1, 1]
            elif len(shape) == 1:
                pass
            elif len(shape) == 3:
                if (
                    output_[0] in op_types
                    and str(op_types[output_[0]]) in _SEQUENCE_LAYERS_REGISTRY
                ):
                    # onnx shape: (Seq,B,C)
                    shape = [shape[2]]
            elif len(shape) == 4:  # (B,C,H,W) --> (C,H,W)
                shape = shape[1:]
            else:
                shape = None  # output shape need not be specified for CoreML.
            if shape is None:
                features.append((str(output_[0]), shape))
            else:
                features.append((str(output_[0]), datatypes.Array(*shape)))
    return features


def _check_unsupported_ops(
    nodes, disable_coreml_rank5_mapping=False
):  # type: (...) -> None
    unsupported_op_types = []  # type: List[Text]
    for node in nodes:

        if disable_coreml_rank5_mapping:
            if (
                node.op_type not in _ONNX_NODE_REGISTRY_ND
                and node.op_type not in unsupported_op_types
            ):
                unsupported_op_types.append(node.op_type)
            continue

        if (
            node.op_type not in _ONNX_NODE_REGISTRY
            and node.op_type not in unsupported_op_types
        ):
            unsupported_op_types.append(node.op_type)

    coreml_3_rerun_message = ""
    if not disable_coreml_rank5_mapping:
        coreml_3_rerun_message = (
            "\nPlease try converting again by providing the additonal argument, "
            "minimum_ios_deployment_target=13"
            " and making sure you have the latest coremltools package"
        )
    if len(unsupported_op_types) > 0:
        raise NotImplementedError(
            "Unsupported ONNX ops of type: %s %s"
            % (",".join(unsupported_op_types), coreml_3_rerun_message)
        )


def _update_multiarray_to_float32(
    feature,  # type: Any
):  # type : (...) -> None
    if feature.type.HasField("multiArrayType"):
        feature.type.multiArrayType.dataType = ft.ArrayFeatureType.FLOAT32


def _update_multiarray_to_int32(
    feature,  # type: Any
):  # type : (...) -> None
    if feature.type.HasField("multiArrayType"):
        feature.type.multiArrayType.dataType = ft.ArrayFeatureType.INT32


def _transform_coreml_dtypes(
    builder,  # type : NeuralNetworkBuilder
    inputs,  # type: List[EdgeInfo]
    outputs,  # type: List[EdgeInfo]
):
    # type: (...) -> None

    """ Make sure ONNX input/output data types are mapped to the equivalent CoreML types
    """
    for i, input_ in enumerate(inputs):
        onnx_type = input_[1]
        if onnx_type == TensorProto.FLOAT:
            _update_multiarray_to_float32(builder.spec.description.input[i])
        elif onnx_type == TensorProto.DOUBLE:
            continue
        elif onnx_type == TensorProto.INT32 or onnx_type == TensorProto.INT64:
            _update_multiarray_to_int32(builder.spec.description.input[i])
        elif onnx_type == TensorProto.BOOL:
            _update_multiarray_to_float32(builder.spec.description.input[i])
        else:
            raise TypeError("Input must be of of type FLOAT, DOUBLE, INT32 or INT64")

    for i, output_ in enumerate(outputs):
        onnx_type = output_[1]
        if onnx_type == TensorProto.FLOAT:
            _update_multiarray_to_float32(builder.spec.description.output[i])
        elif onnx_type == TensorProto.DOUBLE:
            continue
        elif onnx_type == TensorProto.INT32 or onnx_type == TensorProto.INT64:
            _update_multiarray_to_int32(builder.spec.description.output[i])
        elif onnx_type == TensorProto.BOOL:
            _update_multiarray_to_float32(builder.spec.description.output[i])
        else:
            raise TypeError("Output must be of of type FLOAT, DOUBLE, INT32 or INT64")


def _convert_multiarray_output_to_image(
    spec,  # type: Any
    feature_name,  # type: Text
    is_bgr=False,  # type: bool
):
    # type: (...) -> None
    for output in spec.description.output:
        if output.name != feature_name:
            continue
        if output.type.WhichOneof("Type") != "multiArrayType":
            raise ValueError("{} is not a multiarray type".format(output.name,))
        array_shape = tuple(output.type.multiArrayType.shape)
        if len(array_shape) == 2:
            height, width = array_shape
            output.type.imageType.colorSpace = ft.ImageFeatureType.ColorSpace.Value(
                "GRAYSCALE"
            )
        else:
            if len(array_shape) == 4:
                if array_shape[0] != 1:
                    raise ValueError(
                        "Shape {} is not supported for image output".format(
                            array_shape,
                        )
                    )
                array_shape = array_shape[1:]

            channels, height, width = array_shape

            if channels == 1:
                output.type.imageType.colorSpace = ft.ImageFeatureType.ColorSpace.Value(
                    "GRAYSCALE"
                )
            elif channels == 3:
                if is_bgr:
                    output.type.imageType.colorSpace = ft.ImageFeatureType.ColorSpace.Value(
                        "BGR"
                    )
                else:
                    output.type.imageType.colorSpace = ft.ImageFeatureType.ColorSpace.Value(
                        "RGB"
                    )
            else:
                raise ValueError(
                    "Channel Value {} is not supported for image output".format(
                        channels,
                    )
                )

        output.type.imageType.width = width
        output.type.imageType.height = height


def _set_deprocessing(
    is_grayscale,  # type: bool
    builder,  # type: NeuralNetworkBuilder
    deprocessing_args,  # type: Dict[Text, Any]
    input_name,  # type: Text
    output_name,  # type: Text
):
    # type: (...) -> None
    is_bgr = deprocessing_args.get("is_bgr", False)

    image_scale = deprocessing_args.get("image_scale", 1.0)

    if is_grayscale:
        gray_bias = deprocessing_args.get("gray_bias", 0.0)
        W = np.array([image_scale])
        b = np.array([gray_bias])
    else:
        W = np.array([image_scale, image_scale, image_scale])

        red_bias = deprocessing_args.get("red_bias", 0.0)
        green_bias = deprocessing_args.get("green_bias", 0.0)
        blue_bias = deprocessing_args.get("blue_bias", 0.0)

        if not is_bgr:
            b = np.array([red_bias, green_bias, blue_bias,])
        else:
            b = np.array([blue_bias, green_bias, red_bias,])
    builder.add_scale(
        name=input_name,
        W=W,
        b=b,
        has_bias=True,
        shape_scale=W.shape,
        shape_bias=b.shape,
        input_name=input_name,
        output_name=output_name,
    )


def _prepare_onnx_graph(
    graph, transformers, onnx_ir_version
):  # type: (Graph, Iterable[Transformer]) -> Graph
    graph_ = Graph.from_onnx(graph, onnx_ir_version)
    if DEBUG:
        plot_graph(graph_, graph_img_path="/tmp/graph_raw.pdf")
    graph_ = graph_.transformed(transformers)
    if DEBUG:
        plot_graph(graph_, graph_img_path="/tmp/graph_opt.pdf")
    return graph_


def convert(
    model,  # type: Union[onnx.ModelProto, Text]
    mode=None,  # type: Optional[Text]
    image_input_names=[],  # type: Sequence[Text]
    preprocessing_args={},  # type: Dict[Text, Any]
    image_output_names=[],  # type: Sequence[Text]
    deprocessing_args={},  # type: Dict[Text, Any]
    class_labels=None,  # type: Union[Text, Iterable[Text], None]
    predicted_feature_name="classLabel",  # type: Text
    add_custom_layers=False,  # type: bool
    custom_conversion_functions={},  # type: Dict[Text, Any]
    onnx_coreml_input_shape_map={},  # type: Dict[Text, List[int,...]]
    minimum_ios_deployment_target="12",
):
    # type: (...) -> MLModel
    """
    Convert ONNX model to CoreML.
    
    Parameters
    ----------
    model:
        An ONNX model with parameters loaded in the ONNX package, or path to file
        with models.
        
    mode: 'classifier', 'regressor' or None
    
        Mode of the converted coreml model:
        
        * ``'classifier'``: a NeuralNetworkClassifier spec will be constructed.
        * ``'regressor'``: a NeuralNetworkRegressor spec will be constructed.
        
    preprocessing_args:
        The ``'is_bgr'``, ``'red_bias'``, ``'green_bias'``, ``'blue_bias'``, ``'gray_bias'``,
        and ``'image_scale'`` keys have the same meaning as the pre-processing arguments for
        `NeuralNetworkBuilder <https://coremltools.readme.io/reference/modelsneural_network>`_.
    
    deprocessing_args:
        Same as ``'preprocessing_args'`` but for de-processing.
    
    class_labels:
        * As a string, it represents the name of the file which contains
          the classification labels (one per line).
        * As a list of strings, it represents a list of categories that map
          the index of the output of a neural network to labels in a classifier.
    
    predicted_feature_name:
        Name of the output feature for the class labels exposed in the Core ML
        model (applies to classifiers only). Defaults to ``'classLabel'``.
    
    add_custom_layers: bool
        Flag to turn on additional custom CoreML layers for unsupported ONNX ops or
    	attributes within a supported op.
    
    custom_conversion_functions: dict()
        * A dictionary with keys corresponding to the names/types of ONNX ops and values as 
          functions taking an object of the ``coreml-tools`` class:
          ``'NeuralNetworkBuilder'``, ``'Graph'`` (see ``onnx-coreml/_graph.Graph``),
          ``'Node'`` (see ``onnx-coreml/_graph.Node``), and 
          ``'ErrorHandling'`` (see ``onnx-coreml/_error_utils.ErrorHandling``).
        * This custom conversion function gets full control and responsibility for 
          converting a given ONNX op.
        * The function returns nothing and is responsible for adding an equivalent CoreML
          layer via ``'NeuralNetworkBuilder'``.
    
    onnx_coreml_input_shape_map: dict() (Optional) 
        * A dictionary with keys corresponding to the model input names.
        * Values are a list of integers that specify how the shape of the input is mapped
          to CoreML.
        * Convention used for CoreML shapes is ``0: Sequence``, ``1: Batch``,
          ``2: channel``, ``3: height``, ``4: width``. For example, an input of rank 2
          could be mapped as ``[3,4]`` (H,W) or ``[1,2]`` (B,C), and so on. This is
          ignored if ``minimum_ios_deployment_target`` is set to ``13``.
    
    minimum_ios_deployment_target: str
        Target Deployment iOS Version (default: ``'12'``). Supported iOS version options:
        ``'11.2'``, ``'12'``, ``'13'``. CoreML model produced by the converter will be
        compatible with the iOS version specified in this argument. For example, if
        ``minimum_ios_deployment_target = '12'``, the converter would utilize only CoreML
        features released up to version iOS12 (equivalent to macOS 10.14, watchOS 5, and
        so on). iOS 11.2 (CoreML 0.8) does not support ``resize_bilinear`` and
        ``crop_resize`` layers. See `supported v0.8 features <https://github.com/apple/coremltools/releases/tag/v0.8>`_.
        iOS 12 (CoreML 2.0), see `supported v2.0 features <https://github.com/apple/coremltools/releases/tag/v2.0>`_.
        iSO 13 (CoreML 3.0), see `supported v3.0 features <https://github.com/apple/coremltools/releases/tag/3.0-beta6>`_.
    
    
    Returns
    -------
    model: A coreml model.
    """
    if not _HAS_ONNX:
        raise ModuleNotFoundError("Missing ONNX package.")

    if isinstance(model, Text):
        onnx_model = onnx.load(model)
    elif isinstance(model, onnx.ModelProto):
        onnx_model = model
    else:
        raise TypeError("Model must be file path to .onnx file or onnx loaded model")

    if not SupportedVersion.ios_support_check(minimum_ios_deployment_target):
        raise TypeError(
            "{} not supported. Please provide one of target iOS: {}",
            minimum_ios_deployment_target,
            SupportedVersion.get_supported_ios(),
        )

    global USE_SHAPE_MAPPING
    disable_coreml_rank5_mapping = False
    if SupportedVersion.is_nd_array_supported(minimum_ios_deployment_target):
        disable_coreml_rank5_mapping = True

    if disable_coreml_rank5_mapping:
        USE_SHAPE_MAPPING = False
    else:
        USE_SHAPE_MAPPING = True

    """
    First, apply a few optimizations to the ONNX graph,
    in preparation for conversion to CoreML.
    """

    # Using Dummy transformation to conditionally disable certain transformation
    class DummyTransformation(object):
        def __call__(self, graph):
            return graph

    transformers = [
        ConstantsToInitializers(),
        ShapeOpRemover(),
        ConstantRemover(),
        CastOpRemover(),
        PaddingOpRemover(),
        ReshapeInitTensorFuser(),
        DropoutRemover(),
        DeadCodeElimination(),
        ConvAddFuser(),
        BNBroadcastedMulFuser(),
        BNBroadcastedAddFuser(),
        ReshapeTransposeReshape_pattern1(),
        PixelShuffleFuser(),
        AddModelInputsOutputs()
        if not disable_coreml_rank5_mapping
        else DummyTransformation(),
        ConstantFillToInitializers(),
    ]  # type: Iterable[Transformer]

    onnx_model = onnx.shape_inference.infer_shapes(onnx_model)
    graph = _prepare_onnx_graph(onnx_model.graph, transformers, onnx_model.ir_version)

    """
    Check for ImageScalar nodes in ONNX, this will indicate whether input image preprocessing needs
    to be added to the CoreML graph or not.
    """
    # are there ImageScaler nodes in the Graph?
    # If yes then add the info from it to the "preprocessing_args" dictionary, if the dictionary is not
    # already provided by the user
    if not bool(preprocessing_args):
        for node in graph.nodes:
            if node.op_type == "ImageScaler":
                inp_name = node.inputs[0]
                scale = node.attrs.get("scale", 1.0)
                bias = node.attrs.get("bias", [0, 0, 0])
                if not (len(bias) == 1 or len(bias) == 3):
                    continue
                if "image_scale" in preprocessing_args:
                    preprocessing_args["image_scale"][inp_name] = scale
                else:
                    preprocessing_args["image_scale"] = {inp_name: scale}
                if len(bias) == 3:
                    for i, color in enumerate(["red", "green", "blue"]):
                        if color + "_bias" in preprocessing_args:
                            preprocessing_args[color + "_bias"][inp_name] = bias[i]
                        else:
                            preprocessing_args[color + "_bias"] = {inp_name: bias[i]}
                else:
                    if "gray_bias" in preprocessing_args:
                        preprocessing_args["gray_bias"][inp_name] = bias[0]
                    else:
                        preprocessing_args["gray_bias"] = {inp_name: bias[0]}
                if inp_name not in image_input_names:
                    image_input_names.append(inp_name)  # type: ignore

    # remove all ImageScaler ops
    graph = graph.transformed([ImageScalerRemover()])

    """
    Gather information (name, shape) for model inputs and outputs
    This information is then used to initialize the neural network builder object of coremltools.
    The builder object is later used to add layers to the CoreML model.
    """

    # Make CoreML input and output features by gathering shape info and
    # interpreting it for CoreML
    input_features = _make_coreml_input_features(
        graph, onnx_coreml_input_shape_map, disable_coreml_rank5_mapping
    )
    if len(image_output_names) > 0:
        output_features = _make_coreml_output_features(
            graph,
            forceShape=True,
            disable_coreml_rank5_mapping=disable_coreml_rank5_mapping,
        )
    else:
        output_features = _make_coreml_output_features(
            graph, disable_coreml_rank5_mapping=disable_coreml_rank5_mapping
        )

    builder = NeuralNetworkBuilder(
        input_features,
        output_features,
        mode=mode,
        disable_rank5_shape_mapping=disable_coreml_rank5_mapping,
    )

    # TODO: To be removed once, auto-downgrading of spec version is enabled
    builder.spec.specificationVersion = SupportedVersion.get_specification_version(
        minimum_ios_deployment_target
    )

    """
    Set CoreML input,output types (float, double, int) same as onnx types, if supported
    """
    _transform_coreml_dtypes(builder, graph.inputs, graph.outputs)

    """what follows is some book-keeping to support outputs of type image.
    """

    is_deprocess_bgr_only = (len(deprocessing_args) == 1) and (
        "is_bgr" in deprocessing_args
    )
    add_deprocess = (
        (len(image_output_names) > 0)
        and (len(deprocessing_args) > 0)
        and (not is_deprocess_bgr_only)
    )

    if add_deprocess:
        mapping = {}
        for f in output_features:
            output_name = f[0]
            mapping[output_name] = graph.get_unique_edge_name(output_name)
        graph = OutputRenamer(mapping)(graph)

    if len(image_input_names) > 0:
        builder.set_pre_processing_parameters(
            image_input_names=image_input_names,
            is_bgr=preprocessing_args.get("is_bgr", False),
            red_bias=preprocessing_args.get("red_bias", 0.0),
            green_bias=preprocessing_args.get("green_bias", 0.0),
            blue_bias=preprocessing_args.get("blue_bias", 0.0),
            gray_bias=preprocessing_args.get("gray_bias", 0.0),
            image_scale=preprocessing_args.get("image_scale", 1.0),
        )

    preprocessing_args.clear()

    if len(image_output_names) > 0:
        for f in output_features:
            f_name = f[0]
            if f_name in image_output_names:
                is_bgr = deprocessing_args.get("is_bgr", False)
                _convert_multiarray_output_to_image(builder.spec, f_name, is_bgr=is_bgr)

    """
    Iterate through all the ONNX ops and translate them to CoreML layers, one by one.
    """

    """
    before proceeding to start the layer translation process,
    check whether there is an op in the ONNX graph, whose translation function is not yet
    implemented in the converter or which is not supported in the CoreML framework. If so,
    raise an error before starting the process.
    (if the user desires to add a custom layer then this check is not required)
    """
    if not add_custom_layers:
        _check_unsupported_ops(graph.nodes, disable_coreml_rank5_mapping)

    """
    ErrorHandling is a generic class, useful to store a variety of parameters during the conversion process
    """
    err = ErrorHandling(add_custom_layers, custom_conversion_functions)

    for i, node in enumerate(graph.nodes):
        print(
            "%d/%d: Converting Node Type %s" % (i + 1, len(graph.nodes), node.op_type)
        )
        if disable_coreml_rank5_mapping:
            _convert_node_nd(builder, node, graph, err)
        else:
            _add_const_inputs_if_required(builder, node, graph, err)
            _convert_node(builder, node, graph, err)

    if DEBUG:
        plot_graph(
            graph,
            graph_img_path="/tmp/after_conversion.pdf",
            show_coreml_mapped_shapes=not disable_coreml_rank5_mapping,
        )

    if add_deprocess:
        for f in output_features:
            output_name = f[0]
            if output_name not in image_output_names:
                continue
            output_shape = f[1].dimensions
            if len(output_shape) == 2 or output_shape[0] == 1:
                is_grayscale = True
            elif output_shape[0] == 3:
                is_grayscale = False
            else:
                raise ValueError("Output must be RGB image or Grayscale")
            _set_deprocessing(
                is_grayscale,
                builder,
                deprocessing_args,
                mapping[output_name],
                output_name,
            )

    if class_labels is not None:
        if isinstance(class_labels, Text):
            labels = [
                l.strip() for l in open(class_labels).readlines()
            ]  # type: Sequence[Text]
        elif isinstance(class_labels, list):
            labels = class_labels
        else:
            raise TypeError(
                "synset variable of unknown type. Type found: {}. \
                Expected either string or list of strings.".format(
                    type(class_labels),
                )
            )

        builder.set_class_labels(
            class_labels=labels, predicted_feature_name=predicted_feature_name
        )

    def _add_informative_description(feature, raise_error=True):
        if feature.type.WhichOneof("Type") == "multiArrayType":
            if (
                feature.name in graph.onnx_coreml_shape_mapping
                and feature.name in graph.shape_dict
            ):
                mapp = graph.onnx_coreml_shape_mapping[feature.name]
                onnx_shape = graph.shape_dict[feature.name]
                if raise_error:
                    assert len(mapp) == len(onnx_shape), "Something wrong in shape"
                if len(mapp) == len(onnx_shape):
                    shape = []
                    for i in range(5):
                        if i in mapp:
                            shape += [int(onnx_shape[mapp.index(i)])]
                        else:
                            shape += [1]
                    msg = "MultiArray of shape {}. The first and second dimensions correspond to sequence and batch size, respectively".format(
                        str(tuple(shape))
                    )
                    feature.shortDescription += msg

    optional_input_names = []
    for tup in graph.optional_inputs:
        optional_input_names.append(tup[0])
    optional_output_names = []
    for tup in graph.optional_outputs:
        optional_output_names.append(tup[0])

    # add description for inputs and outputs shapes
    remove_input_id = []
    for i, input_ in enumerate(builder.spec.description.input):
        if input_.name not in optional_input_names:
            if not disable_coreml_rank5_mapping:
                _add_informative_description(input_)
        else:
            remove_input_id.append(i)
    remove_output_id = []
    for i, output_ in enumerate(builder.spec.description.output):
        if output_.name not in optional_output_names:
            if not disable_coreml_rank5_mapping:
                _add_informative_description(output_, raise_error=False)
        else:
            remove_output_id.append(i)

    for index in sorted(remove_input_id, reverse=True):
        del builder.spec.description.input[index]
    for index in sorted(remove_output_id, reverse=True):
        del builder.spec.description.output[index]

    if len(graph.optional_inputs) > 0 or len(graph.optional_outputs):
        builder.add_optionals(graph.optional_inputs, graph.optional_outputs)

    # Check for specification version and target ios compatibility
    if (
        minimum_ios_deployment_target == "11.2"
        and builder.spec.WhichOneof("Type") == "neuralNetwork"
    ):
        nn_spec = builder.spec.neuralNetwork
        for layer in nn_spec.layers:
            if (
                layer.WhichOneof("layer") == "resizeBilinear"
                or layer.WhichOneof("layer") == "cropResize"
            ):
                raise TypeError(
                    "{} not supported with target iOS 11.2 please provide higher target iOS".format(
                        layer.WhichOneof("layer")
                    )
                )

    # Optimize ML Model Spec
    ml_model_passes = [remove_disconnected_layers, transform_conv_crop]
    for opt in ml_model_passes:
        opt(builder.spec)

    print("Translation to CoreML spec completed. Now compiling the CoreML model.")
    try:
        if DEBUG:
            import coremltools

            coremltools.models.utils.save_spec(
                builder.spec, "/tmp/node_model_raw_spec.mlmodel"
            )
            from coremltools.models.neural_network.printer import print_network_spec

            print_network_spec(builder.spec, style="coding")
        mlmodel = MLModel(builder.spec)
    except RuntimeError as e:
        raise ValueError("Compilation failed: {}".format(str(e)))
    print("Model Compilation done.")

    # print information about all ops for which custom layers have been added
    if len(err.custom_layer_nodes) > 0:
        print("\n")
        print(
            "Custom layers have been added to the CoreML model "
            "corresponding to the following ops in the onnx model: "
        )
        for i, node in enumerate(err.custom_layer_nodes):
            input_info = []
            for input_ in node.inputs:
                input_info.append(
                    (
                        str(input_),
                        graph.shape_dict.get(input_, str("Shape not available")),
                    )
                )
            output_info = []
            for output_ in node.outputs:
                output_info.append(
                    (
                        str(output_),
                        graph.shape_dict.get(output_, str("Shape not available")),
                    )
                )
            print(
                "{}/{}: op type: {}, op input names and shapes: {}, op output names and shapes: {}".format(
                    i + 1,
                    len(err.custom_layer_nodes),
                    node.op_type,
                    str(input_info),
                    str(output_info),
                )
            )

    mlmodel.user_defined_metadata[_METADATA_VERSION] = ct_version
    mlmodel.user_defined_metadata[_METADATA_SOURCE] = "onnx=={0}".format(
        onnx.__version__
    )
    return mlmodel



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import logging

from coremltools.converters.mil.mil.types.symbolic import is_symbolic
from coremltools.converters.mil.mil import Program, Function
from coremltools.converters.mil.testing_utils import compare_backend
from coremltools.converters.mil.testing_reqs import ct

UNK_VARIADIC = "*s_unk"
UNK_SYM = "s_unk"


def run_compare_builder(
    build,
    input_placeholders,
    input_values,
    expected_output_types=None,
    expected_outputs=None,
    use_cpu_only=False,
    frontend_only=False,
    backend="nn_proto",
    atol=1e-04,
    rtol=1e-05,
    inputs=None,
    also_compare_shapes=False,
):
    """
    Inputs:
        - build: python function taking input of Vars and returning Var or
          list[Var]. Each input argument in build must match a key in
          input_values / input_placeholders.

        - input_placeholders: str -> placeholder. It may not be an empty
                              dict as MLModel doesn't support function with
                              no input.

        - input_values: str -> np.array or PIL.Image. Keys must match those in
          input_placeholders.

        - expected_output_types: list[(shape, builtin_type)] or (shape,
          builtin_type).  None skips type inference validation.

        - expected_outputs: list[np.array] or np.array. Required iff
          frontend_only == False

        - frontend_only: True to test up to proto generation.

        - inputs: type of inputs (either None (defaults to tensor) or [ct.ImageType])
    """
    if not isinstance(expected_output_types, list):
        expected_output_types = [expected_output_types]

    if expected_outputs is not None and not isinstance(expected_outputs, list):
        expected_outputs = [expected_outputs]

    prog = Program()
    with Function(input_placeholders) as ssa_func:
        output_vars = build(**ssa_func.inputs)
        if isinstance(output_vars, tuple):
            output_vars = list(output_vars)
        elif not isinstance(output_vars, list):
            output_vars = [output_vars]
        ssa_func.set_outputs(output_vars)
        prog.add_function("main", ssa_func)

    # get output names for output_vars
    output_names = [x.name for x in output_vars]

    # Validate type inference
    msg = (
        "Provided expected outputs types {} should match number of output"
        + " variables {}"
    )
    assert_msg = msg.format(len(expected_output_types), len(output_vars))
    assert len(output_vars) == len(expected_output_types), assert_msg

    for out_var, s in zip(output_vars, expected_output_types):
        if out_var.dtype != s[-1]:
            raise ValueError(
                "Output {} type: expect {}, got {}. Program:\n{}".format(
                    out_var.name, s[-1].__type_info__(),
                    out_var.dtype.__type_info__(), prog
                )
            )
        if UNK_VARIADIC in s[:-1]:
            msg = "Skip type checking for UNK_VARIADIC. Output shape: {} vs expected shape: {}"
            logging.debug(msg.format(out_var.shape, s[:-1]))
            continue
        expected_shape = s[:-1]
        msg = "Output {} shape: expect {}, got {}. Program:\n{}".format(
            out_var.name, expected_shape, out_var.shape, prog
        )
        # No more variadic here.
        if len(out_var.shape) != len(expected_shape):
            raise ValueError(msg)
        # replace UNK_SYM in out_var.shape.
        output_shape = [
            0 if es == UNK_SYM else os for os, es in zip(out_var.shape, expected_shape)
        ]
        expected_shape = [0 if es == UNK_SYM else es for es in expected_shape]
        # convert float etc to int.
        output_shape = [i if is_symbolic(i) else int(i) for i in output_shape]
        expected_shape = [i if is_symbolic(i) else int(i) for i in expected_shape]
        if output_shape != expected_shape:
            raise ValueError(msg)

    mlmodel = ct.convert(prog, source="mil", convert_to=backend, inputs=inputs)

    if frontend_only:
        return

    if expected_outputs:
        assert len(output_vars) == len(expected_outputs), (
            "Provided expected_outputs {}"
            " should match number of output"
            " variables {}".format(len(expected_outputs), len(output_vars))
        )

        expected_outputs = {
            name: val for name, val in zip(output_names, expected_outputs)
        }

    compare_backend(
        mlmodel=mlmodel,
        input_key_values=input_values,
        expected_outputs=expected_outputs,
        use_cpu_only=use_cpu_only,
        atol=atol,
        rtol=rtol,
        also_compare_shapes=also_compare_shapes,
    )



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.testing_utils import (
    assert_op_count_match,
    assert_model_is_valid,
    get_op_types_in_program,
    apply_pass_and_basic_check,
)
import unittest
import pytest

import numpy as np

np.random.seed(1984)


class PadConvOptimizationPass(unittest.TestCase):
    """
    Input graph:
    input -----> pad -----> transpose -----> conv -----> transpose ---> out

    Output graph:
    input -----> transpose -----> pad ----> conv -----> transpose ----> out
    """

    def test_simple_direct_output(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 16, 20, 24))])
        def prog(x):
            x = mb.pad(x=x, pad=[0,0,1,1,1,1,0,0])
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x = mb.conv(x=x, weight=np.random.random([24,24,3,3]), pad_type="valid")
            x = mb.transpose(x=x, perm=[0, 2, 3, 1])
            return x

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::pad_conv_connect"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["pad", "transpose", "conv", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["transpose", "pad", "conv", "transpose"])
        assert_model_is_valid(
            prog,
            {"x": (1, 16, 20, 24)},
            expected_output_shapes={block.outputs[0].name: (1, 16, 20, 24)},
        )

    """
    Input graph:
    input -----> pad -----> transpose -----> conv -----> transpose ---> out
                  |
                  |
                  --------> transpose -----> conv -----> transpose ---> out

    Output graph:
    input ---------> transpose -----> pad -----> conv -----> transpose ---> out
             |
             |
             ------> transpose -----> pad -----> conv -----> transpose ---> out

    """

    def test_pad_transposed_forked_conv(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 16, 20, 24))])
        def prog(x):
            pad = mb.pad(x=x, pad=[0,0,1,1,1,1,0,0])
            x = mb.transpose(x=pad, perm=[0, 3, 1, 2])
            x = mb.conv(x=x, weight=np.random.random([24,24,3,3]), pad_type="valid")
            x = mb.transpose(x=x, perm=[0, 2, 3, 1])
            y = mb.transpose(x=pad, perm=[0, 3, 1, 2])
            y = mb.conv(x=y, weight=np.random.random([24,24,3,3]), pad_type="valid")
            y = mb.transpose(x=y, perm=[0, 2, 3, 1])
            return x, y

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::pad_conv_connect"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["pad", "transpose", "conv", "transpose", "transpose", "conv", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["transpose", "pad", "conv", "transpose", "transpose", "pad", "conv", "transpose"])
        assert_model_is_valid(
            prog,
            {"x": (1, 16, 20, 24)},
            expected_output_shapes={block.outputs[0].name: (1, 16, 20, 24),
                                    block.outputs[1].name: (1, 16, 20, 24)},
        )

    """
    Input graph:
    input -----> pad -----> transpose -----> conv -----> transpose ---> out
                  |
                  |
                  ---------> out

    Output graph:
    No change.
    """

    def test_pad_output(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 16, 20, 24))])
        def prog(x):
            pad = mb.pad(x=x, pad=[0,0,1,1,1,1,0,0])
            x = mb.transpose(x=pad, perm=[0, 3, 1, 2])
            x = mb.conv(x=x, weight=np.random.random([24,24,3,3]), pad_type="valid")
            x = mb.transpose(x=x, perm=[0, 2, 3, 1])
            return x, pad

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::pad_conv_connect"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["pad", "transpose", "conv", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["pad", "transpose", "conv", "transpose"])
        assert_model_is_valid(
            prog,
            {"x": (1, 16, 20, 24)},
            expected_output_shapes={block.outputs[0].name: (1, 16, 20, 24),
                                    block.outputs[1].name: (1, 18, 22, 24)},
        )




#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.testing_utils import (
    assert_model_is_valid,
    get_op_types_in_program,
    apply_pass_and_basic_check,
)
from coremltools.converters.mil.mil.passes.pass_registry import PASS_REGISTRY
import copy
import pytest
import itertools

import numpy as np


@pytest.mark.parametrize("op_type, pos, val", itertools.product(['add', 'mul', 'floor_div', 'pow', 'real_div', 'sub'], ['x', 'y'], [0, 1, [0, 0, 0, 0], [1, 1, 1, 1]]))
def test_elementwise_elimination(op_type, pos, val):
    if 'div' in op_type and np.prod(val) == 0:
        return
    if 'pow' in op_type and (val != 0 or val != 1):
        return

    test_op = getattr(mb, op_type)

    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        if pos == "x":
            r1 = test_op(x=val, y=x)
        else:
            r1 = test_op(x=x, y=val)
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    original_program = [op_type, "relu"]
    new_program = original_program
    if op_type in {'add'}:
        if val == 0 or val == [0, 0, 0, 0]:
            new_program = ["relu"]
    elif op_type in {'mul'}:
        if val == 1 or val == [1, 1, 1, 1]:
            new_program = ["relu"]
    elif op_type in {'pow', 'real_div', 'floor_div'}:
        if pos == 'y' and (val == 1 or val == [1, 1, 1, 1]):
            new_program = ["relu"]
    elif op_type in {'sub'}:
        if pos == 'y' and (val == 0 or val == [0, 0, 0, 0]):
            new_program = ["relu"]
            
    assert get_op_types_in_program(prev_prog) == original_program
    assert get_op_types_in_program(prog) == new_program
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )

def test_elementwise_broadcast():

    @mb.program(input_specs=[mb.TensorSpec(shape=[4])])
    def prog(x):
        r1 = mb.add(x=x, y=[[0, 0, 0, 0], [0, 0, 0, 0]])
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    original_program = ["add", "relu"]

    assert get_op_types_in_program(prev_prog) == original_program
    assert get_op_types_in_program(prog) == original_program
    assert_model_is_valid(
        prog,
        {"x": [4]},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )

def test_reshape_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.reshape(x=x, shape=[1, 8])
        r2 = mb.reshape(x=r1, shape=[1, 8])
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["reshape", "reshape", "relu"]
    assert get_op_types_in_program(prog) == ["reshape", "relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (1, 8)},
    )


def test_oneway_split_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.split(x=x, num_splits=1, axis=-1) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["split", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )


def test_full_split_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.split(x=x, split_sizes=[4], axis=-1) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["split", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )


def test_slicebysize_full_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.slice_by_size(x=x, begin=[0, 0], size=[2, 4]) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["slice_by_size", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )


def test_slicebysize_to_end_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.slice_by_size(x=x, begin=[0, 0], size=[-1, -1]) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["slice_by_size", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )


def test_slicebyindex_full_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.slice_by_index(x=x, begin=[0, 0], end=[2, 4]) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["slice_by_index", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )


@pytest.mark.parametrize("begin_mask, end_mask",
                         itertools.product(itertools.product([True, False],[True, False]),
                                           itertools.product([True, False],[True, False])))
def test_slicebyindex_mask_elimination(begin_mask, end_mask):
    @mb.program(input_specs=[mb.TensorSpec(shape=(4, 4))])
    def prog(x):
        begin = [1, 1]
        end = [1, 1]
        for i in range(2):
            if not begin_mask[i]:
                begin[i] = 0
            if not end_mask[i]:
                end[i] = 4
        r1 = mb.slice_by_index(x=x, begin=begin, end=end, begin_mask=begin_mask, end_mask=end_mask) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["slice_by_index", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (4, 4)},
        expected_output_shapes={block.outputs[0].name: (4, 4)},
    )


def test_pad_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.pad(x=x, pad=[0, 0, 0, 0]) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["pad", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )


def test_keep_pad():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.pad(x=x, pad=[4, 4, 2, 2]) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["pad", "relu"]
    assert get_op_types_in_program(prog) == ["pad", "relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (10, 8)},
    )


def test_tile_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.tile(x=x, reps=[1, 1]) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["tile", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )


def test_keep_tile():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.tile(x=x, reps=[2, 2]) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["tile", "relu"]
    assert get_op_types_in_program(prog) == ["tile", "relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (4, 8)},
    )


def test_upsample_nearest_neighbor_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(3, 2, 4))])
    def prog(x):
        r1 = mb.upsample_nearest_neighbor(x=x) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["upsample_nearest_neighbor", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (3, 2, 4)},
        expected_output_shapes={block.outputs[0].name: (3, 2, 4)},
    )


def test_upsample_bilinear_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(3, 2, 4))])
    def prog(x):
        r1 = mb.upsample_bilinear(x=x) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["upsample_bilinear", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (3, 2, 4)},
        expected_output_shapes={block.outputs[0].name: (3, 2, 4)},
    )


def test_resize_bilinear_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(3, 2, 4))])
    def prog(x):
        r1 = mb.resize_bilinear(x=x, target_size_height=2, target_size_width=4) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["resize_bilinear", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (3, 2, 4)},
        expected_output_shapes={block.outputs[0].name: (3, 2, 4)},
    )


def test_crop_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(3, 2, 4))])
    def prog(x):
        r1 = mb.crop(x=x, crop_height=[0, 0], crop_width=[0, 0]) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["crop", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (3, 2, 4)},
        expected_output_shapes={block.outputs[0].name: (3, 2, 4)},
    )


def test_linear_elimination():
    @mb.program(input_specs=[mb.TensorSpec(shape=(2, 4))])
    def prog(x):
        r1 = mb.linear_activation(x=x, alpha=1.0, beta=0.0) 
        return mb.relu(x=r1)

    prev_prog, prev_block, block = apply_pass_and_basic_check(
        prog, "common::noop_elimination"
    )
    assert get_op_types_in_program(prev_prog) == ["linear_activation", "relu"]
    assert get_op_types_in_program(prog) == ["relu"]
    assert_model_is_valid(
        prog,
        {"x": (2, 4)},
        expected_output_shapes={block.outputs[0].name: (2, 4)},
    )





from __future__ import absolute_import as _
from __future__ import division as _
from __future__ import print_function as _
from __future__ import unicode_literals as _

import pytest

onnx = pytest.importorskip("onnx")

import unittest
import numpy as np
import numpy.testing as npt  # type: ignore

from coremltools._deps import _HAS_ONNX, MSG_ONNX_NOT_FOUND

if _HAS_ONNX:
    import onnx
    from onnx import helper, numpy_helper, TensorProto

    from coremltools.converters.onnx import convert
    from coremltools.converters.onnx._graph import Graph
    from coremltools.converters.onnx._transformers import (
        ConvAddFuser,
        DropoutRemover,
        ImageScalerRemover,
    )
    from ._test_utils import (
        _onnx_create_model,
        _test_onnx_model,
        _conv_pool_output_size,
        _random_array,
    )


@unittest.skipUnless(_HAS_ONNX, MSG_ONNX_NOT_FOUND)
class ConvAddFuserTest(unittest.TestCase):
    def test_fuse_conv_without_bias(self):  # type: () -> None
        kernel_shape = (3, 2)
        strides = (2, 3)
        pads = (4, 2, 4, 2)
        dilations = (1, 2)
        group = 1
        weight = numpy_helper.from_array(_random_array((16, 3, 3, 2)), name="weight")

        input_shape = (1, 3, 224, 224)
        output_size = _conv_pool_output_size(
            input_shape, dilations, kernel_shape, pads, strides
        )

        output_shape = (1, int(weight.dims[0]), output_size[0], output_size[1])

        inputs = [("input0", input_shape)]
        outputs = [("output0", output_shape, TensorProto.FLOAT)]

        conv = helper.make_node(
            "Conv",
            inputs=[inputs[0][0], "weight"],
            outputs=["conv_output"],
            dilations=dilations,
            group=group,
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

        b = _random_array((int(weight.dims[0]),))
        bias = numpy_helper.from_array(b, name="bias")

        add = helper.make_node(
            "Add",
            inputs=[conv.output[0], "bias"],
            outputs=[outputs[0][0]],
            broadcast=1,
            axis=1,
        )

        model = _onnx_create_model([conv, add], inputs, outputs, [weight, bias])
        graph_ = Graph.from_onnx(model.graph, onnx_ir_version=5)
        fused_graph = graph_.transformed([ConvAddFuser()])

        self.assertEqual(len(fused_graph.nodes), 1)
        node = fused_graph.nodes[0]
        self.assertEqual(len(node.inputs), 3)
        npt.assert_equal(node.input_tensors[node.inputs[2]], b)
        self.assertEqual(fused_graph.nodes[0].outputs[0], outputs[0][0])

    def test_fuse_conv_with_bias(self):  # type: () -> None
        kernel_shape = (3, 2)
        strides = (2, 3)
        pads = (4, 2, 4, 2)
        dilations = (1, 2)
        group = 1
        weight = numpy_helper.from_array(_random_array((16, 3, 3, 2)), name="weight")
        b = _random_array((int(weight.dims[0]),))
        bias = numpy_helper.from_array(b, name="bias")

        input_shape = (1, 3, 224, 224)
        output_size = _conv_pool_output_size(
            input_shape, dilations, kernel_shape, pads, strides
        )

        output_shape = (1, int(weight.dims[0]), output_size[0], output_size[1])

        inputs = [("input0", input_shape)]
        outputs = [("output0", output_shape, TensorProto.FLOAT)]

        conv = helper.make_node(
            "Conv",
            inputs=[inputs[0][0], "weight", "bias"],
            outputs=["conv_output"],
            dilations=dilations,
            group=group,
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

        add = helper.make_node(
            "Add",
            inputs=[conv.output[0], "bias"],
            outputs=[outputs[0][0]],
            broadcast=1,
            axis=1,
        )

        model = _onnx_create_model([conv, add], inputs, outputs, [weight, bias])
        graph_ = Graph.from_onnx(model.graph, onnx_ir_version=5)
        fused_graph = graph_.transformed([ConvAddFuser()])

        self.assertEqual(len(fused_graph.nodes), 1)
        node = fused_graph.nodes[0]
        self.assertEqual(len(node.inputs), 3)
        npt.assert_equal(node.input_tensors[node.inputs[2]], b * 2)
        self.assertEqual(fused_graph.nodes[0].outputs[0], outputs[0][0])


@unittest.skipUnless(_HAS_ONNX, MSG_ONNX_NOT_FOUND)
class NodeRemoverTests(unittest.TestCase):
    def test_dropout_remover(self):  # type: () -> None
        inputs = [("input", (1, 3, 50, 50))]
        outputs = [("out", (1, 5, 50, 50), TensorProto.FLOAT)]
        weight = numpy_helper.from_array(_random_array((5, 3, 1, 1)), name="weight")
        conv = helper.make_node(
            "Conv",
            inputs=["input", "weight"],
            outputs=["conv_output"],
            kernel_shape=(1, 1),
            strides=(1, 1),
        )
        drop = helper.make_node(
            "Dropout", inputs=["conv_output"], outputs=["drop_output"],
        )
        exp = helper.make_node("Exp", inputs=["drop_output"], outputs=["out"])

        onnx_model = _onnx_create_model([conv, drop, exp], inputs, outputs)

        graph = Graph.from_onnx(onnx_model.graph, onnx_ir_version=5)
        new_graph = graph.transformed([DropoutRemover()])
        self.assertEqual(len(graph.nodes), 3)
        self.assertEqual(len(new_graph.nodes), 2)
        self.assertEqual(new_graph.nodes[0].inputs[0], "input")
        self.assertEqual(new_graph.nodes[1].inputs[0], new_graph.nodes[0].outputs[0])
        self.assertEqual(new_graph.nodes[1].outputs[0], "out")

    def test_image_scaler_remover(self):  # type: () -> None
        inputs = [("input", (1, 3, 50, 50))]
        outputs = [("out", (1, 3, 50, 50), TensorProto.FLOAT)]

        im_scaler = helper.make_node(
            "ImageScaler",
            inputs=["input"],
            outputs=["scaler_out"],
            bias=[10, -6, 20],
            scale=3.0,
        )

        exp = helper.make_node("Exp", inputs=["scaler_out"], outputs=["out"])

        onnx_model = _onnx_create_model([im_scaler, exp], inputs, outputs)

        graph = Graph.from_onnx(onnx_model.graph, onnx_ir_version=5)
        new_graph = graph.transformed([ImageScalerRemover()])
        self.assertEqual(len(graph.nodes), 2)
        self.assertEqual(len(new_graph.nodes), 1)
        self.assertEqual(new_graph.nodes[0].inputs[0], "input")
        self.assertEqual(new_graph.nodes[0].outputs[0], "out")

        coreml_model = convert(onnx_model)
        spec = coreml_model.get_spec()

        self.assertEqual(spec.neuralNetwork.preprocessing[0].scaler.channelScale, 3.0)
        self.assertEqual(spec.neuralNetwork.preprocessing[0].scaler.blueBias, 20.0)
        self.assertEqual(spec.neuralNetwork.preprocessing[0].scaler.greenBias, -6.0)
        self.assertEqual(spec.neuralNetwork.preprocessing[0].scaler.redBias, 10.0)

    def test_multiple_image_scaler(self):  # type : () -> None
        inputs = [("input_color", (1, 3, 10, 10)), ("input_gray", (1, 1, 10, 10))]
        outputs = [("out", (1, 4, 10, 10), TensorProto.FLOAT)]

        im_scaler1 = helper.make_node(
            "ImageScaler",
            inputs=["input_color"],
            outputs=["scaler_out_1"],
            bias=[10, -6, 20],
            scale=3.0,
        )

        im_scaler2 = helper.make_node(
            "ImageScaler",
            inputs=["input_gray"],
            outputs=["scaler_out_2"],
            bias=[-13],
            scale=5.0,
        )

        concat = helper.make_node(
            "Concat", inputs=["scaler_out_1", "scaler_out_2"], outputs=["out"], axis=1
        )

        onnx_model = _onnx_create_model(
            [im_scaler1, im_scaler2, concat], inputs, outputs
        )

        spec = convert(onnx_model).get_spec()
        self.assertEqual(len(spec.neuralNetwork.layers), 1)
        self.assertEqual(len(spec.neuralNetwork.preprocessing), 2)
        self.assertEqual(spec.neuralNetwork.preprocessing[0].scaler.channelScale, 3.0)
        self.assertEqual(spec.neuralNetwork.preprocessing[0].scaler.blueBias, 20.0)
        self.assertEqual(spec.neuralNetwork.preprocessing[0].scaler.greenBias, -6.0)
        self.assertEqual(spec.neuralNetwork.preprocessing[0].scaler.redBias, 10.0)
        self.assertEqual(spec.neuralNetwork.preprocessing[1].scaler.channelScale, 5.0)
        self.assertEqual(spec.neuralNetwork.preprocessing[1].scaler.grayBias, -13.0)


@unittest.skipUnless(_HAS_ONNX, MSG_ONNX_NOT_FOUND)
class PixelShuffleFuserTest(unittest.TestCase):
    def test_pixel_shuffle(self):  # type: () -> None
        scale_factor = 2
        input_shape = (1, 8, 2, 2)
        output_shape = (
            input_shape[0],
            int(input_shape[1] / (scale_factor ** 2)),
            input_shape[2] * scale_factor,
            input_shape[3] * scale_factor,
        )

        inputs = [("input0", input_shape)]
        outputs = [("output0", output_shape, TensorProto.FLOAT)]

        shape1 = [
            output_shape[0],
            output_shape[1],
            scale_factor,
            scale_factor,
            input_shape[2],
            input_shape[3],
        ]

        shape1 = numpy_helper.from_array(np.asarray(shape1), name="shape1")
        shape2 = numpy_helper.from_array(np.asarray(list(output_shape)), name="shape2")

        node_0 = helper.make_node(
            "Reshape", inputs=[inputs[0][0], "shape1"], outputs=["node0"],
        )
        node_1 = helper.make_node(
            "Transpose", inputs=["node0"], outputs=["node1"], perm=[0, 1, 4, 2, 5, 3]
        )
        node_2 = helper.make_node(
            "Reshape", inputs=["node1", "shape2"], outputs=[outputs[0][0]],
        )
        model = _onnx_create_model(
            [node_0, node_1, node_2], inputs, outputs, initializer=[shape1, shape2]
        )
        _test_onnx_model(model, decimal=7)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest
import numpy as np
import os
import shutil
import tempfile
import itertools
import coremltools
from coremltools._deps import _HAS_KERAS_TF, _HAS_KERAS2_TF
from coremltools.models.utils import _macos_version, _is_macos
import pytest

if _HAS_KERAS_TF or _HAS_KERAS2_TF:
    from keras.models import Sequential
    from keras.layers import LSTM, GRU, SimpleRNN
    from coremltools.converters import keras as keras_converter


def _get_mlkit_model_from_path(model, model_path):
    from coremltools.converters import keras as keras_converter

    print("converting")
    model = keras_converter.convert(model, ["data"], ["output"])
    return model


def generate_input(dim0, dim1, dim2):
    input_data = np.random.rand(dim0, dim1, dim2).astype(
        "f"
    )  # astype() should be removed
    return input_data


def valid_params(params):
    """Checks if this combination of parameters is allowed by Keras"""
    return not (params["input_dims"][1] == 1 and params["unroll"])


class RecurrentLayerTest(unittest.TestCase):
    """
    Base class for recurrent layer tests. Masking param not included here
    """

    def setUp(self):
        self.params_dict = dict(
            input_dims=[[1, 1, 1], [1, 1, 5], [1, 1, 10]],  # [1, x > 1, y] not added
            output_dim=[1, 5, 10, 20],
            stateful=[False, True],
            go_backwards=[False],  # True],
            unroll=[False, True],
            return_sequences=[False, True],
            activation=["sigmoid", "tanh", "hard_sigmoid", "linear"],
        )
        self.base_layer_params = list(itertools.product(*self.params_dict.values()))


class SimpleRNNLayer(RecurrentLayerTest):
    """
    Class for testing single RNN layer
    """

    def setUp(self):
        super(SimpleRNNLayer, self).setUp()
        self.simple_rnn_params_dict = dict(
            dropout=[{"dropout_W": 0.0, "dropout_U": 0.0}],
            regularizer=[
                {"W_regularizer": None, "U_regularizer": None, "b_regularizer": None}
            ],
        )
        self.rnn_layer_params = list(
            itertools.product(*self.simple_rnn_params_dict.values())
        )

    def _test_rnn_layer(self, keras_major_version, limit=None):
        i = 0
        layer_name = str(SimpleRNN).split(".")[3].split("'>")[0]
        numerical_err_models = []
        shape_err_models = []
        params = list(itertools.product(self.base_layer_params, self.rnn_layer_params))
        np.random.shuffle(params)
        params = [
            param
            for param in params
            if valid_params(dict(zip(self.params_dict.keys(), param[0])))
        ]
        for base_params, rnn_params in params[:limit]:
            base_params = dict(zip(self.params_dict.keys(), base_params))
            rnn_params = dict(zip(self.simple_rnn_params_dict.keys(), rnn_params))
            input_data = generate_input(
                base_params["input_dims"][0],
                base_params["input_dims"][1],
                base_params["input_dims"][2],
            )
            model = Sequential()
            settings = dict(
                activation=base_params["activation"],
                return_sequences=base_params["return_sequences"],
                go_backwards=base_params["go_backwards"],
                unroll=base_params["unroll"],
            )
            if keras_major_version == 2:
                model.add(
                    SimpleRNN(
                        base_params["output_dim"],
                        input_shape=base_params["input_dims"][1:],
                        dropout=rnn_params["dropout"]["dropout_U"],
                        recurrent_dropout=rnn_params["dropout"]["dropout_W"],
                        kernel_regularizer=rnn_params["regularizer"]["W_regularizer"],
                        recurrent_regularizer=rnn_params["regularizer"][
                            "U_regularizer"
                        ],
                        bias_regularizer=rnn_params["regularizer"]["b_regularizer"],
                        **settings
                    )
                )
            else:
                model.add(
                    SimpleRNN(
                        base_params["output_dim"],
                        input_length=base_params["input_dims"][1],
                        input_dim=base_params["input_dims"][2],
                        dropout_U=rnn_params["dropout"]["dropout_U"],
                        dropout_W=rnn_params["dropout"]["dropout_W"],
                        W_regularizer=rnn_params["regularizer"]["W_regularizer"],
                        U_regularizer=rnn_params["regularizer"]["U_regularizer"],
                        b_regularizer=rnn_params["regularizer"]["b_regularizer"],
                        **settings
                    )
                )
            model_dir = tempfile.mkdtemp()
            keras_model_path = os.path.join(model_dir, "keras.h5")
            coreml_model_path = os.path.join(model_dir, "keras.mlmodel")
            model.save_weights(keras_model_path)
            mlkitmodel = _get_mlkit_model_from_path(model, coreml_model_path)
            if _is_macos() and _macos_version() >= (10, 13):
                keras_preds = model.predict(input_data).flatten()
                input_data = np.transpose(input_data, [1, 0, 2])
                coreml_preds = mlkitmodel.predict({"data": input_data})[
                    "output"
                ].flatten()
                try:
                    self.assertEquals(coreml_preds.shape, keras_preds.shape)
                except AssertionError:
                    print(
                        "Shape error:\nbase_params: {}\nkeras_preds.shape: {}\ncoreml_preds.shape: {}".format(
                            base_params, keras_preds.shape, coreml_preds.shape
                        )
                    )
                    shape_err_models.append(base_params)
                    shutil.rmtree(model_dir)
                    i += 1
                    continue
                try:
                    for idx in range(0, len(coreml_preds)):
                        relative_error = (
                            coreml_preds[idx] - keras_preds[idx]
                        ) / coreml_preds[idx]
                        self.assertAlmostEqual(relative_error, 0, places=2)
                except AssertionError:
                    print(
                        "Assertion error:\nbase_params: {}\nkeras_preds: {}\ncoreml_preds: {}".format(
                            base_params, keras_preds, coreml_preds
                        )
                    )
                    numerical_err_models.append(base_params)
            shutil.rmtree(model_dir)
            i += 1

        self.assertEquals(
            shape_err_models, [], msg="Shape error models {}".format(shape_err_models)
        )
        self.assertEquals(
            numerical_err_models,
            [],
            msg="Numerical error models {}".format(numerical_err_models),
        )

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    @pytest.mark.slow
    def test_keras1_rnn_layer_stress(self):
        self._test_rnn_layer(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_rnn_layer(self):
        self._test_rnn_layer(keras_major_version=1, limit=10)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_rnn_layer_stress(self):
        self._test_rnn_layer(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_rnn_layer(self):
        self._test_rnn_layer(keras_major_version=2, limit=10)


class LSTMLayer(RecurrentLayerTest):
    """
    Class for testing single RNN layer
    """

    def setUp(self):
        super(LSTMLayer, self).setUp()
        self.lstm_params_dict = dict(
            dropout=[{"dropout_W": 0.0, "dropout_U": 0.0}],
            regularizer=[
                {"W_regularizer": None, "U_regularizer": None, "b_regularizer": None}
            ],
        )
        self.lstm_layer_params = list(
            itertools.product(*self.lstm_params_dict.values())
        )

    def _test_lstm_layer(self, keras_major_version, limit=None):
        i = 0
        numerical_err_models = []
        shape_err_models = []
        params = list(itertools.product(self.base_layer_params, self.lstm_layer_params))
        np.random.shuffle(params)
        params = [
            param
            for param in params
            if valid_params(dict(zip(self.params_dict.keys(), param[0])))
        ]
        for base_params, lstm_params in params[:limit]:
            base_params = dict(zip(self.params_dict.keys(), base_params))
            lstm_params = dict(zip(self.lstm_params_dict.keys(), lstm_params))
            input_data = generate_input(
                base_params["input_dims"][0],
                base_params["input_dims"][1],
                base_params["input_dims"][2],
            )
            model = Sequential()
            settings = dict(
                activation=base_params["activation"],
                return_sequences=base_params["return_sequences"],
                go_backwards=base_params["go_backwards"],
                unroll=base_params["unroll"],
            )
            if keras_major_version == 2:
                model.add(
                    LSTM(
                        base_params["output_dim"],
                        input_shape=base_params["input_dims"][1:],
                        recurrent_dropout=lstm_params["dropout"]["dropout_U"],
                        dropout=lstm_params["dropout"]["dropout_W"],
                        kernel_regularizer=lstm_params["regularizer"]["W_regularizer"],
                        recurrent_regularizer=lstm_params["regularizer"][
                            "U_regularizer"
                        ],
                        bias_regularizer=lstm_params["regularizer"]["b_regularizer"],
                        **settings
                    )
                )
            else:
                model.add(
                    LSTM(
                        base_params["output_dim"],
                        input_length=base_params["input_dims"][1],
                        input_dim=base_params["input_dims"][2],
                        dropout_U=lstm_params["dropout"]["dropout_U"],
                        dropout_W=lstm_params["dropout"]["dropout_W"],
                        W_regularizer=lstm_params["regularizer"]["W_regularizer"],
                        U_regularizer=lstm_params["regularizer"]["U_regularizer"],
                        b_regularizer=lstm_params["regularizer"]["b_regularizer"],
                        **settings
                    )
                )
            model_dir = tempfile.mkdtemp()
            keras_model_path = os.path.join(model_dir, "keras.h5")
            coreml_model_path = os.path.join(model_dir, "keras.mlmodel")
            model.save_weights(keras_model_path)
            mlkitmodel = _get_mlkit_model_from_path(model, coreml_model_path)
            if _is_macos() and _macos_version() >= (10, 13):
                keras_preds = model.predict(input_data).flatten()
                input_data = np.transpose(input_data, [1, 0, 2])
                coreml_preds = mlkitmodel.predict({"data": input_data})[
                    "output"
                ].flatten()
                try:
                    self.assertEquals(coreml_preds.shape, keras_preds.shape)
                except AssertionError:
                    print(
                        "Shape error:\nbase_params: {}\nkeras_preds.shape: {}\ncoreml_preds.shape: {}".format(
                            base_params, keras_preds.shape, coreml_preds.shape
                        )
                    )
                    shape_err_models.append(base_params)
                    shutil.rmtree(model_dir)
                    i += 1
                    continue
                try:
                    for idx in range(0, len(coreml_preds)):
                        relative_error = (
                            coreml_preds[idx] - keras_preds[idx]
                        ) / coreml_preds[idx]
                        self.assertAlmostEqual(relative_error, 0, places=2)
                except AssertionError:
                    print(
                        "Assertion error:\nbase_params: {}\nkeras_preds: {}\ncoreml_preds: {}".format(
                            base_params, keras_preds, coreml_preds
                        )
                    )
                    numerical_err_models.append(base_params)
            shutil.rmtree(model_dir)
            i += 1

        self.assertEquals(
            shape_err_models, [], msg="Shape error models {}".format(shape_err_models)
        )
        self.assertEquals(
            numerical_err_models,
            [],
            msg="Numerical error models {}".format(numerical_err_models),
        )

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    @pytest.mark.slow
    def test_keras1_lstm_layer_stress(self):
        self._test_lstm_layer(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_lstm_layer(self):
        self._test_lstm_layer(keras_major_version=1, limit=10)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_lstm_layer_stress(self):
        self._test_lstm_layer(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_lstm_layer(self):
        self._test_lstm_layer(keras_major_version=2, limit=10)


class GRULayer(RecurrentLayerTest):
    """
    Class for testing GRU layer
    """

    def setUp(self):
        super(GRULayer, self).setUp()
        self.gru_params_dict = dict(
            dropout=[{"dropout_W": 0.0, "dropout_U": 0.0}],
            regularizer=[
                {"W_regularizer": None, "U_regularizer": None, "b_regularizer": None}
            ],
        )
        self.gru_layer_params = list(itertools.product(*self.gru_params_dict.values()))

    def _test_gru_layer(self, keras_major_version, limit=None):
        i = 0
        numerical_err_models = []
        shape_err_models = []
        params = list(itertools.product(self.base_layer_params, self.gru_layer_params))
        np.random.shuffle(params)
        params = [
            param
            for param in params
            if valid_params(dict(zip(self.params_dict.keys(), param[0])))
        ]
        for base_params, gru_params in params[:limit]:
            base_params = dict(zip(self.params_dict.keys(), base_params))
            gru_params = dict(zip(self.gru_params_dict.keys(), gru_params))
            input_data = generate_input(
                base_params["input_dims"][0],
                base_params["input_dims"][1],
                base_params["input_dims"][2],
            )
            model = Sequential()
            settings = dict(
                activation=base_params["activation"],
                return_sequences=base_params["return_sequences"],
                go_backwards=base_params["go_backwards"],
                unroll=base_params["unroll"],
            )
            if keras_major_version == 2:
                model.add(
                    GRU(
                        base_params["output_dim"],
                        input_shape=base_params["input_dims"][1:],
                        recurrent_dropout=gru_params["dropout"]["dropout_U"],
                        dropout=gru_params["dropout"]["dropout_W"],
                        kernel_regularizer=gru_params["regularizer"]["W_regularizer"],
                        recurrent_regularizer=gru_params["regularizer"][
                            "U_regularizer"
                        ],
                        bias_regularizer=gru_params["regularizer"]["b_regularizer"],
                        **settings
                    )
                )
            else:
                model.add(
                    GRU(
                        base_params["output_dim"],
                        input_length=base_params["input_dims"][1],
                        input_dim=base_params["input_dims"][2],
                        dropout_U=gru_params["dropout"]["dropout_U"],
                        dropout_W=gru_params["dropout"]["dropout_W"],
                        W_regularizer=gru_params["regularizer"]["W_regularizer"],
                        U_regularizer=gru_params["regularizer"]["U_regularizer"],
                        b_regularizer=gru_params["regularizer"]["b_regularizer"],
                        **settings
                    )
                )
            model_dir = tempfile.mkdtemp()
            keras_model_path = os.path.join(model_dir, "keras.h5")
            coreml_model_path = os.path.join(model_dir, "keras.mlmodel")
            model.save_weights(keras_model_path)
            mlkitmodel = _get_mlkit_model_from_path(model, coreml_model_path)
            if _is_macos() and _macos_version() >= (10, 13):
                keras_preds = model.predict(input_data).flatten()
                input_data = np.transpose(input_data, [1, 0, 2])
                coreml_preds = mlkitmodel.predict({"data": input_data})[
                    "output"
                ].flatten()
                try:
                    self.assertEquals(coreml_preds.shape, keras_preds.shape)
                except AssertionError:
                    print(
                        "Shape error:\nbase_params: {}\nkeras_preds.shape: {}\ncoreml_preds.shape: {}".format(
                            base_params, keras_preds.shape, coreml_preds.shape
                        )
                    )
                    shape_err_models.append(base_params)
                    shutil.rmtree(model_dir)
                    i += 1
                    continue
                try:
                    for idx in range(0, len(coreml_preds)):
                        relative_error = (
                            coreml_preds[idx] - keras_preds[idx]
                        ) / coreml_preds[idx]
                        self.assertAlmostEqual(relative_error, 0, places=2)
                except AssertionError:
                    print(
                        "Assertion error:\nbase_params: {}\nkeras_preds: {}\ncoreml_preds: {}".format(
                            base_params, keras_preds, coreml_preds
                        )
                    )
                    numerical_err_models.append(base_params)
            shutil.rmtree(model_dir)
            i += 1

        self.assertEquals(
            shape_err_models, [], msg="Shape error models {}".format(shape_err_models)
        )
        self.assertEquals(
            numerical_err_models,
            [],
            msg="Numerical error models {}".format(numerical_err_models),
        )

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    @pytest.mark.slow
    def test_keras1_gru_layer_stress(self):
        self._test_gru_layer(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_gru_layer(self):
        self._test_gru_layer(keras_major_version=1, limit=10)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_gru_layer_stress(self):
        self._test_gru_layer(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_gru_layer(self):
        self._test_gru_layer(keras_major_version=2, limit=10)



import unittest

import numpy as np

import coremltools
from coremltools.models import datatypes, MLModel
from coremltools.models.neural_network import NeuralNetworkBuilder
from coremltools.models.neural_network.quantization_utils import (
    _convert_array_to_nbit_quantized_bytes,
    quantize_weights,
)
from coremltools.models.utils import _macos_version, _is_macos
from coremltools.converters.mil.backend.nn.op_mapping import to_py_type

MIN_MACOS_VERSION_REQUIRED = (10, 13)
LAYERS_10_14_MACOS_VERSION = (10, 14)
LAYERS_10_15_MACOS_VERSION = (10, 15)


@unittest.skipIf(
    not _is_macos() or _macos_version() < LAYERS_10_15_MACOS_VERSION,
    "Only supported on macOS 10.15+",
)
class ControlFlowCorrectnessTest(unittest.TestCase):
    @classmethod
    def setup_class(cls):
        pass

    def runTest():
        pass

    def _test_model(self, model, input_dict, output_ref, delta=1e-2):
        preds = model.predict(input_dict)
        for name in output_ref:
            ref_val = output_ref[name]
            val = preds[name]
            self.assertTrue(np.allclose(val, ref_val, rtol=delta))

    def test_simple_branch(self):
        """ Test a simple if-else branch network
        """
        input_features = [("data", datatypes.Array(3)), ("cond", datatypes.Array(1))]
        output_features = [("output", None)]

        builder_top = NeuralNetworkBuilder(
            input_features, output_features, disable_rank5_shape_mapping=True
        )
        layer = builder_top.add_branch("branch_layer", "cond")

        builder_ifbranch = NeuralNetworkBuilder(
            input_features=None,
            output_features=None,
            spec=None,
            nn_spec=layer.branch.ifBranch,
        )
        builder_ifbranch.add_elementwise(
            "mult_layer",
            input_names=["data"],
            output_name="output",
            mode="MULTIPLY",
            alpha=10,
        )
        builder_elsebranch = NeuralNetworkBuilder(
            input_features=None,
            output_features=None,
            spec=None,
            nn_spec=layer.branch.elseBranch,
        )
        builder_elsebranch.add_elementwise(
            "add_layer",
            input_names=["data"],
            output_name="output",
            mode="ADD",
            alpha=10,
        )
        coremltools.models.utils.save_spec(
            builder_top.spec, "/tmp/simple_branch.mlmodel"
        )
        mlmodel = MLModel(builder_top.spec)

        # True branch case
        input_dict = {
            "data": np.array(range(1, 4), dtype="float"),
            "cond": np.array([1], dtype="float"),
        }
        output_ref = {"output": input_dict["data"] * 10}
        self._test_model(mlmodel, input_dict, output_ref)

        # False branch case
        input_dict["cond"] = np.array([0], dtype="float")
        output_ref["output"] = input_dict["data"] + 10
        self._test_model(mlmodel, input_dict, output_ref)

    def test_simple_loop_fixed_iterations(self):
        input_features = [("data", datatypes.Array(1))]
        output_features = [("output", None)]

        builder_top = NeuralNetworkBuilder(
            input_features, output_features, disable_rank5_shape_mapping=True
        )
        builder_top.add_copy("copy_1", input_name="data", output_name="output")

        loop_layer = builder_top.add_loop("loop_layer")
        loop_layer.loop.maxLoopIterations = 5
        builder_body = NeuralNetworkBuilder(
            input_features=None,
            output_features=None,
            spec=None,
            nn_spec=loop_layer.loop.bodyNetwork,
        )
        builder_body.add_elementwise(
            "add", input_names=["output"], output_name="x", mode="ADD", alpha=2
        )

        builder_body.add_copy("copy_2", input_name="x", output_name="output")
        coremltools.models.utils.save_spec(
            builder_top.spec, "/tmp/simple_loop_fixed_iterations.mlmodel"
        )
        mlmodel = MLModel(builder_top.spec)

        # True branch case
        input_dict = {"data": np.array([0], dtype="float")}
        output_ref = {"output": np.array([10], dtype="float")}
        self._test_model(mlmodel, input_dict, output_ref)


@unittest.skipUnless(
    _is_macos() and _macos_version() >= LAYERS_10_14_MACOS_VERSION,
    "Only supported on macOS 10.14+",
)
class BasicNumericCorrectnessTest_1014NewLayers(unittest.TestCase):
    def build_quant_conv_layer(
        self,
        W=None,
        quantization_type="linear",
        nbits=8,
        quant_scale=None,
        quant_bias=None,
        quant_lut=None,
        output_channels=2,
    ):
        input_features = [("data", datatypes.Array(1, 2, 2))]
        output_features = [("out", datatypes.Array(2, 1, 1))]
        builder = NeuralNetworkBuilder(input_features, output_features)
        builder.add_convolution(
            name="conv",
            kernel_channels=1,
            output_channels=output_channels,
            height=2,
            width=2,
            stride_height=1,
            stride_width=1,
            border_mode="valid",
            groups=1,
            W=W,
            b=None,
            has_bias=False,
            input_name="data",
            output_name="out",
            quantization_type=quantization_type,
            nbits=nbits,
            quant_scale=quant_scale,
            quant_bias=quant_bias,
            quant_lut=quant_lut,
        )
        return MLModel(builder.spec)

    def test_linear_quant_convolution_8bit(self):
        W = np.ones((2, 2, 1, 2), dtype=np.uint8)
        W[:, :, :, 1] = 2
        mlmodel = self.build_quant_conv_layer(
            W=W.flatten().tobytes(),
            quantization_type="linear",
            nbits=8,
            quant_scale=[4.0],
            quant_bias=[-2.0],
        )
        data = np.ones((1, 2, 2))
        data_dict = {"data": data}
        out = mlmodel.predict(data_dict, useCPUOnly=True)["out"]
        expected_out = np.reshape(np.array([8, 24]), (2, 1, 1))
        self.assertTrue(np.allclose(out, expected_out))

    def test_linear_quant_convolution_8bit_vector_scalebias(self):
        W = np.ones((2, 2, 1, 2), dtype=np.uint8)
        W[:, :, :, 1] = 2
        mlmodel = self.build_quant_conv_layer(
            W=W.flatten().tobytes(),
            quantization_type="linear",
            nbits=8,
            quant_scale=[4.0, 5.0],
            quant_bias=[-2.0, 1.0],
        )
        data = np.ones((1, 2, 2))
        data_dict = {"data": data}
        out = mlmodel.predict(data_dict, useCPUOnly=True)["out"]
        expected_out = np.reshape(np.array([8, 44]), (2, 1, 1))
        self.assertTrue(np.allclose(out, expected_out))

    @unittest.skip("<rdar://problem/68866645> Investigate numerical discrepancy during quantization in CoreML")
    def test_linear_quant_convolution_8bit_float_scale_and_bias(self):
        W = np.array(([[[[1, 248], [248, 248]]]]), dtype=np.uint8)
        mlmodel = self.build_quant_conv_layer(
            W=W.flatten().tobytes(),
            quantization_type="linear",
            nbits=8,
            quant_scale=[15.346457],
            quant_bias=[-3913.3464],
            output_channels=1,
        )
        data = np.ones((1, 2, 2))
        data_dict = {"data": data}
        out = mlmodel.predict(data_dict, useCPUOnly=True)["out"]
        # Output should be equal to: (scale*(1+248+248+248)+(4*bias))
        expected_out = np.reshape(np.array([-4220.275]), (1, 1, 1, 1, 1))
        self.assertTrue(np.allclose(out, expected_out))

    def test_lut_quant_convolution_2bit(self):
        W = np.zeros((2, 2, 1, 2), dtype=np.uint8)
        W[:, :, :, 0] = 0
        W[:, :, :, 1] = 2
        W = _convert_array_to_nbit_quantized_bytes(W.flatten(), 2).tobytes()
        mlmodel = self.build_quant_conv_layer(
            W=W, quantization_type="lut", nbits=2, quant_lut=[10.0, 11.0, -3.0, -1.0]
        )
        data = np.ones((1, 2, 2))
        data_dict = {"data": data}
        out = mlmodel.predict(data_dict, useCPUOnly=True)["out"]
        expected_out = np.reshape(np.array([40, -12]), (2, 1, 1))
        self.assertTrue(np.allclose(out, expected_out))

    def test_linear_quant_inner_product_3bit(self):
        W = np.reshape(np.arange(6), (2, 3)).astype(np.uint8)
        input_features = [("data", datatypes.Array(3))]
        output_features = [("probs", None)]
        builder = NeuralNetworkBuilder(input_features, output_features)
        builder.add_inner_product(
            name="ip1",
            W=_convert_array_to_nbit_quantized_bytes(W.flatten(), 3).tobytes(),
            b=None,
            input_channels=3,
            output_channels=2,
            has_bias=False,
            input_name="data",
            output_name="probs",
            quantization_type="linear",
            nbits=3,
            quant_scale=[11.0, 2.0],
            quant_bias=[-2.0, 10.0],
        )
        mlmodel = MLModel(builder.spec)
        data = np.array([1.0, 3.0, 5.0])
        data_dict = {"data": data}
        probs = mlmodel.predict(data_dict)["probs"]
        expected_out = np.array([125, 170])
        self.assertTrue(np.allclose(probs.flatten(), expected_out.flatten()))

    def test_lut_quant_inner_product_1bit(self):
        W = np.zeros((2, 3), dtype=np.uint8)
        W[0, :] = [0, 1, 1]
        W[1, :] = [1, 0, 0]
        input_features = [("data", datatypes.Array(3))]
        output_features = [("probs", None)]
        builder = NeuralNetworkBuilder(input_features, output_features)
        builder.add_inner_product(
            name="ip1",
            W=_convert_array_to_nbit_quantized_bytes(W.flatten(), 1).tobytes(),
            b=None,
            input_channels=3,
            output_channels=2,
            has_bias=False,
            input_name="data",
            output_name="probs",
            quantization_type="lut",
            nbits=1,
            quant_lut=[5.0, -3.0],
        )
        mlmodel = MLModel(builder.spec)
        data = np.array([1.0, 3.0, 5.0])
        data_dict = {"data": data}
        probs = mlmodel.predict(data_dict)["probs"]
        expected_out = np.array([-19, 37])
        self.assertTrue(np.allclose(probs.flatten(), expected_out.flatten()))


@unittest.skipUnless(
    _is_macos() and _macos_version() >= LAYERS_10_15_MACOS_VERSION,
    "Only supported on macOS 10.15+",
)
class BasicNumericCorrectnessTest_1015NewLayers(unittest.TestCase):
    def test_linear_quant_batchedmatmul_5bit(self):
        W = np.zeros((2, 3), dtype=np.uint8)
        W[0, :] = [31, 20, 11]
        W[1, :] = [1, 0, 8]
        quant_scale = np.reshape(np.array([10.0, 2.0, 3.0]), (1, 3))
        quant_bias = np.reshape(np.array([-2.0, -10.0, 6.0]), (1, 3))
        W_unquantized = np.broadcast_to(quant_scale, (2, 3)) * W + np.broadcast_to(
            quant_bias, (2, 3)
        )
        bias = np.array([1.0, 2.0, 3.0])

        input_features = [("data", datatypes.Array(2, 2))]
        output_features = [("out", None)]
        builder = NeuralNetworkBuilder(
            input_features, output_features, disable_rank5_shape_mapping=True
        )
        builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="out",
            weight_matrix_rows=2,
            weight_matrix_columns=3,
            W=_convert_array_to_nbit_quantized_bytes(W.flatten(), 5).tobytes(),
            bias=bias,
            is_quantized_weight=True,
            quantization_type="linear",
            nbits=5,
            quant_scale=quant_scale.flatten(),
            quant_bias=quant_bias.flatten(),
        )
        mlmodel = MLModel(builder.spec)
        data = np.zeros((2, 2), dtype=np.float32)
        data[0, :] = [5, 6]
        data[1, :] = [10, 12]
        data_dict = {"data": data}
        out = mlmodel.predict(data_dict, useCPUOnly=True)["out"]
        expected_out = np.matmul(data, W_unquantized) + bias
        self.assertTrue(out.shape == expected_out.shape)
        self.assertTrue(np.allclose(out.flatten(), expected_out.flatten()))

    def test_linear_quant_batchedmatmul_8bit(self):
        np.random.seed(1988)
        W = np.random.rand(32, 32) * 2.0 - 1
        bias = np.random.rand(32)

        input_features = [("data", datatypes.Array(2, 32))]
        output_features = [("out", None)]
        builder = NeuralNetworkBuilder(
            input_features, output_features, disable_rank5_shape_mapping=True
        )
        builder.add_batched_mat_mul(
            name="batched_matmul",
            input_names=["data"],
            output_name="out",
            weight_matrix_rows=32,
            weight_matrix_columns=32,
            W=W,
            bias=bias,
        )
        mlmodel = MLModel(builder.spec)
        q_mlmodel = quantize_weights(mlmodel, 8)
        q_spec = q_mlmodel.get_spec()
        q_layer = q_spec.neuralNetwork.layers[0].batchedMatmul

        self.assertTrue(len(q_layer.weights.floatValue) == 0)
        self.assertTrue(len(q_layer.weights.rawValue) > 0)

        data = np.random.rand(2, 32)
        data_dict = {"data": data}
        out = q_mlmodel.predict(data_dict, useCPUOnly=True)["out"]
        expected_out = np.matmul(data, W) + bias
        self.assertTrue(out.shape == expected_out.shape)
        self.assertTrue(np.allclose(out.flatten(), expected_out.flatten(), atol=0.1))

    def test_lut_quant_embedding_nd_2bit(self):
        embed_size = 2
        vocab_size = 3
        W = np.zeros((embed_size, vocab_size), dtype=np.uint8)
        W[:, 0] = [1, 0]
        W[:, 1] = [0, 1]
        W[:, 2] = [3, 2]
        bias = np.array([1.0, 2.0])
        quant_lut = np.array([34.0, 12.0, -6.0, 6.0])

        input_features = [("data", datatypes.Array(4, 1))]
        output_features = [("out", None)]
        builder = NeuralNetworkBuilder(
            input_features, output_features, disable_rank5_shape_mapping=True
        )
        builder.add_embedding_nd(
            name="embedding_nd",
            input_name="data",
            output_name="out",
            vocab_size=vocab_size,
            embedding_size=embed_size,
            W=_convert_array_to_nbit_quantized_bytes(W.flatten(), 2).tobytes(),
            b=bias,
            is_quantized_weight=True,
            quantization_type="lut",
            nbits=2,
            quant_lut=quant_lut,
        )

        mlmodel = MLModel(builder.spec)
        data = np.reshape(np.array([2.0, 2.0, 1.0, 0.0]), (4, 1))
        data_dict = {"data": data}
        out = mlmodel.predict(data_dict, useCPUOnly=True)["out"]
        expected_out = np.zeros((4, embed_size), dtype=np.float32)
        expected_out[0, :] = [quant_lut[W[0, 2]], quant_lut[W[1, 2]]] + bias
        expected_out[1, :] = [quant_lut[W[0, 2]], quant_lut[W[1, 2]]] + bias
        expected_out[2, :] = [quant_lut[W[0, 1]], quant_lut[W[1, 1]]] + bias
        expected_out[3, :] = [quant_lut[W[0, 0]], quant_lut[W[1, 0]]] + bias
        self.assertTrue(out.shape == expected_out.shape)
        self.assertTrue(np.allclose(out.flatten(), expected_out.flatten()))

    def test_linear_quant_embedding_7bit(self):
        embed_size = 2
        vocab_size = 3
        W = np.zeros((embed_size, vocab_size), dtype=np.uint8)
        W[:, 0] = [100, 127]
        W[:, 1] = [20, 40]
        W[:, 2] = [90, 1]
        quant_scale = np.reshape(np.array([10.0, 2.0]), (2, 1))
        quant_bias = np.reshape(np.array([-2.0, -10.0]), (2, 1))
        W_unquantized = np.broadcast_to(quant_scale, (2, 3)) * W + np.broadcast_to(
            quant_bias, (2, 3)
        )
        bias = np.reshape(np.array([1.0, 2.0]), (2, 1))
        W_unquantized = W_unquantized + np.broadcast_to(bias, (2, 3))

        input_features = [("data", datatypes.Array(4, 1, 1, 1))]
        output_features = [("out", None)]
        builder = NeuralNetworkBuilder(
            input_features, output_features, disable_rank5_shape_mapping=True
        )
        builder.add_embedding(
            name="embed",
            W=_convert_array_to_nbit_quantized_bytes(W.flatten(), 7).tobytes(),
            b=bias,
            input_dim=vocab_size,
            output_channels=embed_size,
            has_bias=True,
            input_name="data",
            output_name="out",
            is_quantized_weight=True,
            quantization_type="linear",
            nbits=7,
            quant_scale=to_py_type(quant_scale),
            quant_bias=to_py_type(quant_bias),
        )

        mlmodel = MLModel(builder.spec)
        data = np.reshape(np.array([2.0, 2.0, 1.0, 0.0]), (4, 1, 1, 1))
        data_dict = {"data": data}
        out = mlmodel.predict(data_dict, useCPUOnly=True)["out"]
        self.assertTrue(out.shape == (4, embed_size, 1, 1))
        expected_out = np.zeros((4, embed_size), dtype=np.float32)
        expected_out[0, :] = W_unquantized[:, 2].flatten()
        expected_out[1, :] = W_unquantized[:, 2].flatten()
        expected_out[2, :] = W_unquantized[:, 1].flatten()
        expected_out[3, :] = W_unquantized[:, 0].flatten()
        self.assertTrue(np.allclose(out.flatten(), expected_out.flatten()))


@unittest.skipIf(
    not _is_macos() or _macos_version() < (10, 13), "Only supported on macOS 10.13+"
)
class BasicNumericCorrectnessTest(unittest.TestCase):
    def _build_nn_with_one_ip_layer(self):
        input_features = [("data", datatypes.Array(3))]
        output_features = [("out", None)]
        builder = NeuralNetworkBuilder(
            input_features, output_features, disable_rank5_shape_mapping=True
        )
        w = np.random.uniform(-0.5, 0.5, (3, 3))
        builder.add_inner_product(
            name="ip1",
            W=w,
            b=None,
            input_channels=3,
            output_channels=3,
            has_bias=False,
            input_name="input",
            output_name="hidden",
        )
        return builder

    def test_undefined_shape_single_output(self):
        W = np.ones((3, 3))
        input_features = [("data", datatypes.Array(3))]
        output_features = [("probs", None)]
        builder = NeuralNetworkBuilder(input_features, output_features)
        builder.add_inner_product(
            name="ip1",
            W=W,
            b=None,
            input_channels=3,
            output_channels=3,
            has_bias=False,
            input_name="data",
            output_name="probs",
        )
        mlmodel = MLModel(builder.spec)
        data = np.ones((3,))
        data_dict = {"data": data}
        probs = mlmodel.predict(data_dict)["probs"]
        self.assertTrue(np.allclose(probs, np.ones(3) * 3))

    def test_set_input(self):
        builder = self._build_nn_with_one_ip_layer()
        builder.set_input(input_names=["data_renamed"], input_dims=[(2,)])

        self.assertEquals(
            builder.spec.description.input[0].type.multiArrayType.shape[0], 2
        )
        self.assertEquals(builder.spec.description.input[0].name, "data_renamed")

    def test_set_input_fail(self):
        builder = self._build_nn_with_one_ip_layer()

        # fails since input_names and input_dims do not have same size
        with self.assertRaises(ValueError):
            builder.set_input(input_names=["data_1", "data_2"], input_dims=[(3,)])

    def test_set_output(self):
        builder = self._build_nn_with_one_ip_layer()
        builder.set_output(output_names=["out_renamed"], output_dims=[(2,)])

        self.assertEquals(
            builder.spec.description.output[0].type.multiArrayType.shape[0], 2
        )
        self.assertEquals(builder.spec.description.output[0].name, "out_renamed")

    def test_set_output_fail(self):
        builder = self._build_nn_with_one_ip_layer()

        # fails since output_names and output_dims do not have same size
        with self.assertRaises(ValueError):
            builder.set_output(output_names=["out_1", "out_2"], output_dims=[(3,)])

    def test_invalid_image_preprocessing_params(self):
        builder = self._build_nn_with_one_ip_layer()
        image_input_names = ["input1", "input2"]
        with self.assertRaises(ValueError):
            image_scale = {"invalid": 1.0 / 255.0}
            builder.set_pre_processing_parameters(
                image_input_names=image_input_names, image_scale=image_scale
            )
        with self.assertRaises(ValueError):
            red_bias = {"invalid": -1}
            builder.set_pre_processing_parameters(
                image_input_names=image_input_names, red_bias=red_bias
            )
        with self.assertRaises(ValueError):
            blue_bias = {"invalid": -1}
            builder.set_pre_processing_parameters(
                image_input_names=image_input_names, blue_bias=blue_bias
            )
        with self.assertRaises(ValueError):
            green_bias = {"invalid": -1}
            builder.set_pre_processing_parameters(
                image_input_names=image_input_names, green_bias=green_bias
            )
        with self.assertRaises(ValueError):
            gray_bias = {"invalid": -1}
            builder.set_pre_processing_parameters(
                image_input_names=image_input_names, gray_bias=gray_bias
            )
        with self.assertRaises(ValueError):
            is_bgr = {"invalid": False}
            builder.set_pre_processing_parameters(
                image_input_names=image_input_names, is_bgr=is_bgr
            )


@unittest.skipUnless(
    _is_macos() and _macos_version() >= (10, 13), "Only supported on macOS 10.13+"
)
class UseFloatArraytypeTest(unittest.TestCase):
    """Test that the boolean flag `use_float_arraytype` correctly changes the datatype of the
    network's inputs and outputs and produces a spec that the `MLModel` class can call `predict`
    with.
    """

    def _test_use_float_array_helper(self, use_float_arraytype):
        input_features = [("data", datatypes.Array(3))]
        output_features = [("probs", None)]
        builder = NeuralNetworkBuilder(
            input_features=input_features,
            output_features=output_features,
            use_float_arraytype=use_float_arraytype,
        )
        weights = np.ones((3, 3))
        builder.add_inner_product(
            name="ip1",
            W=weights,
            b=None,
            input_channels=3,
            output_channels=3,
            has_bias=False,
            input_name="data",
            output_name="probs",
        )
        spec = builder.spec
        array_feature_type = (
            coremltools.proto.FeatureTypes_pb2.ArrayFeatureType.FLOAT32
            if use_float_arraytype
            else coremltools.proto.FeatureTypes_pb2.ArrayFeatureType.DOUBLE
        )
        for input in spec.description.input:
            self.assertEquals(input.type.multiArrayType.dataType, array_feature_type)
        for output in spec.description.input:
            self.assertEquals(output.type.multiArrayType.dataType, array_feature_type)

        # Assert that the generated spec is functional
        mlmodel = MLModel(spec)
        data = np.ones((3,))
        data_dict = {"data": data}
        try:
            predictions = mlmodel.predict(data_dict)
        except Exception as e:
            self.fail(e)
        self.assertTrue(np.allclose(predictions["probs"], np.ones(3) * 3))

    def test_true_use_float_array(self):
        # Instruct the builder to use the Float32 datatype for inputs and outputs
        self._test_use_float_array_helper(True)

    def test_false_use_float_array(self):
        # Instruct the builder to use its default Double datatype for inputs and outputs
        self._test_use_float_array_helper(False)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import logging


class PassRegistry:
    def __init__(self):
        # str -> func (func takes Program as input and
        # modifies in-place)
        self.passes = {}

    def __getitem__(self, pass_id):
        """
        pass_id (str): namespace::func_name (e.g., 'common::const_elimination')
        """
        if pass_id not in self.passes:
            raise KeyError("Pass {} not found".format(pass_id))
        return self.passes[pass_id]

    def add(self, namespace, pass_func):
        func_name = pass_func.__name__
        pass_id = namespace + "::" + func_name
        logging.debug("Registering pass {}".format(pass_id))
        if pass_id in self.passes:
            msg = "Pass {} already registered."
            raise KeyError(msg.format(pass_id))
        self.passes[pass_id] = pass_func


PASS_REGISTRY = PassRegistry()


def register_pass(namespace):
    """
    namespaces like {'common', 'nn_backend', <other-backends>,
    <other-frontends>}
    """

    def func_wrapper(pass_func):
        PASS_REGISTRY.add(namespace, pass_func)
        return pass_func

    return func_wrapper



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest
import pytest
from coremltools._deps import _HAS_KERAS_TF

if _HAS_KERAS_TF:
    from keras.models import Model
    from keras.layers import Dense, Input, merge
    from coremltools.converters import keras


@unittest.skipIf(not _HAS_KERAS_TF, "Missing keras. Skipping tests.")
@pytest.mark.keras1
class KerasNonSequentialModelTest(unittest.TestCase):
    """
    Unit test class for testing non-sequential Keras models.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading common utilities.
        """

    def test_simple_merge(self):
        """
        Test the following Keras model
               |- dense-|
        dense -|        |- merge - dense
               |- dense-|
        """
        input_tensor = Input(shape=(3,))
        x1 = Dense(4)(input_tensor)
        x2 = Dense(5)(x1)
        x3 = Dense(6)(x1)
        x4 = merge([x2, x3], mode="concat")
        x5 = Dense(7)(x4)

        model = Model(input=[input_tensor], output=[x5])
        input_names = ["data"]
        output_names = ["output"]

        spec = keras.convert(model, input_names, output_names).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertTrue(spec.HasField("neuralNetwork"))

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.input), len(input_names))
        self.assertEqual(
            sorted(input_names), sorted(map(lambda x: x.name, spec.description.input))
        )
        self.assertEqual(len(spec.description.output), len(output_names))
        self.assertEqual(
            sorted(output_names), sorted(map(lambda x: x.name, spec.description.output))
        )

    def test_merge_add(self):
        """
        Test the following Keras model
               |- dense-|
        dense -|        |- merge - dense
               |- dense-|
        """
        input_tensor = Input(shape=(3,))
        x1 = Dense(4)(input_tensor)
        x2 = Dense(5)(x1)
        x3 = Dense(5)(x1)
        x4 = merge([x2, x3], mode="sum")
        x5 = Dense(7)(x4)

        model = Model(input=[input_tensor], output=[x5])
        input_names = ["data"]
        output_names = ["output"]

        spec = keras.convert(model, input_names, output_names).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertTrue(spec.HasField("neuralNetwork"))

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.input), len(input_names))
        self.assertEqual(
            sorted(input_names), sorted(map(lambda x: x.name, spec.description.input))
        )
        self.assertEqual(len(spec.description.output), len(output_names))
        self.assertEqual(
            sorted(output_names), sorted(map(lambda x: x.name, spec.description.output))
        )

    def test_merge_multiply(self):
        """
        Test the following Keras model
               |- dense-|
        dense -|        |- merge - dense
               |- dense-|
        """
        input_tensor = Input(shape=(3,))
        x1 = Dense(4)(input_tensor)
        x2 = Dense(5)(x1)
        x3 = Dense(5)(x1)
        x4 = merge([x2, x3], mode="mul")
        x5 = Dense(7)(x4)

        model = Model(input=[input_tensor], output=[x5])
        input_names = ["data"]
        output_names = ["output"]

        spec = keras.convert(model, input_names, output_names).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertTrue(spec.HasField("neuralNetwork"))

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.input), len(input_names))
        self.assertEqual(
            sorted(input_names), sorted(map(lambda x: x.name, spec.description.input))
        )
        self.assertEqual(len(spec.description.output), len(output_names))
        self.assertEqual(
            sorted(output_names), sorted(map(lambda x: x.name, spec.description.output))
        )



import os
import shutil
import tempfile
import unittest

import numpy as np

import coremltools
import coremltools.models.datatypes as datatypes
from coremltools.models import neural_network as neural_network
from coremltools.models.utils import _macos_version, _is_macos


class SimpleTest(unittest.TestCase):
    def test_fixed_seq_len(self):
        """
        Input has a fixed sequence length.
        (this happens when model is trained using padded sequences, inspiration: https://forums.developer.apple.com/thread/80407)

        (Seq,Batch,C,H,W)
        embedding: input shape (15,1,1,1,1) --> output shape (15,1,32,1,1)
        permute  : input shape (15,1,32,1,1) --> output shape (1,1,32,1,15)
        flatten  : input shape (1,1,32,1,15) --> output shape (1,1,32 * 15,1,1)
        dense    : input shape (1,1,480,1,1) --> output shape (1,1,2,1,1)
        """

        coreml_preds = []
        input_dim = (1, 1, 1)
        output_dim = (
            1,
            1,
            1,
        )  # some random dimensions here: we are going to remove this information later
        input_features = [("data", datatypes.Array(*input_dim))]
        output_features = [("output", datatypes.Array(*output_dim))]
        builder = neural_network.NeuralNetworkBuilder(input_features, output_features)

        # ADD Layers
        builder.add_embedding(
            "embed",
            W=np.random.rand(10, 32),
            b=None,
            input_dim=10,
            output_channels=32,
            has_bias=0,
            input_name="data",
            output_name="embed",
        )
        builder.add_permute(
            "permute", dim=[3, 1, 2, 0], input_name="embed", output_name="permute"
        )
        builder.add_flatten(
            "flatten", mode=0, input_name="permute", output_name="flatten"
        )
        builder.add_inner_product(
            "dense",
            W=np.random.rand(480, 2),
            b=None,
            input_channels=480,
            output_channels=2,
            has_bias=0,
            input_name="flatten",
            output_name="output",
        )

        # Remove output shape by deleting and adding an output
        del builder.spec.description.output[-1]
        output = builder.spec.description.output.add()
        output.name = "output"
        output.type.multiArrayType.dataType = coremltools.proto.FeatureTypes_pb2.ArrayFeatureType.ArrayDataType.Value(
            "DOUBLE"
        )

        # save the model
        model_dir = tempfile.mkdtemp()
        model_path = os.path.join(model_dir, "test_layer.mlmodel")
        coremltools.utils.save_spec(builder.spec, model_path)
        # preprare input and get predictions
        coreml_model = coremltools.models.MLModel(model_path)
        X = np.random.randint(low=0, high=10, size=15)
        X = np.reshape(X, (15, 1, 1, 1, 1)).astype(np.float32)
        coreml_input = {"data": X}
        if _is_macos() and _macos_version() >= (10, 13):
            coreml_preds = coreml_model.predict(coreml_input)["output"]
            self.assertEquals(len(coreml_preds.flatten()), 2)

        if os.path.exists(model_dir):
            shutil.rmtree(model_dir)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

import numpy as np
from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb

child_op_types = ["add", "sub"]


def match_pattern(op):
    if op.op_type == "matmul":
        # find add
        child_ops = op.outputs[0].child_ops
        if len(child_ops) == 1:
            add_op_candidate = list(child_ops)[0]
            if add_op_candidate.op_type in child_op_types:
                return add_op_candidate
    return None


def transpose(v, before_op):
    """
    Transpose the last 2 dims.
    v: Var (must be a tensor)
    """
    perm = list(range(v.rank))
    perm[-2], perm[-1] = perm[-1], perm[-2]
    return mb.transpose(x=v, perm=perm, before_op=before_op)


def try_to_transform(matmul_op, add_op, block):
    if matmul_op.x.val is None and matmul_op.y.val is None:
        # This is a dynamic matmul.
        return False
    if add_op.x.val is None and add_op.y.val is None:
        # This is a dynamic add.
        return False

    x_is_weight = matmul_op.x.val is not None
    if x_is_weight:
        weight, linear_x = matmul_op.x, matmul_op.y
        transpose_weight = matmul_op.transpose_x.val
        transpose_x = matmul_op.transpose_y.val
    else:
        weight, linear_x = matmul_op.y, matmul_op.x
        transpose_weight = matmul_op.transpose_y.val
        transpose_x = matmul_op.transpose_x.val

    if linear_x.rank < 2 or weight.rank != 2:
        # We don't support these cases yet.
        return False

    d_out = weight.shape[1] if not transpose_weight else weight.shape[0]
    bias = add_op.x.val if add_op.x.val is not None else add_op.y.val
    if len(bias.shape) > 1:
        if any([d != 1 for d in bias.shape[:-1]]):
            return  # cannot transform

        # squeeze leading dims of size 1
        bias = np.squeeze(bias)

    if len(bias.shape) != 1 or bias.shape[0] != d_out:
        return  # cannot transform

    if add_op.op_type == "sub":
        bias = -bias
    out_name = add_op.outputs[0].name

    if x_is_weight:
        # If transpose_x == transpose_weight == False:
        # w*x = (x^T w^T)^T = linear(x^T, w)^T
        x_transposed = (
            transpose(linear_x, before_op=matmul_op) if not transpose_x else linear_x
        )
        w_no_transpose = (
            weight if not transpose_weight else transpose(weight, before_op=matmul_op)
        )
        x = mb.linear(
            x=x_transposed, weight=w_no_transpose, bias=bias, before_op=matmul_op
        )
        x = transpose(x, before_op=matmul_op, name=out_name)
    else:
        # If transpose_x == transpose_weight == False
        # x*w = x*(w^T)^T = linear(x, w^T)
        x_no_transpose = (
            transpose(linear_x, before_op=matmul_op) if transpose_x else linear_x
        )
        w_transposed = (
            weight if transpose_weight else transpose(weight, before_op=matmul_op)
        )
        x = mb.linear(
            x=x_no_transpose,
            weight=w_transposed,
            bias=bias,
            before_op=matmul_op,
            name=out_name,
        )

    add_op.enclosing_block.replace_uses_of_var_after_op(
        anchor_op=add_op, old_var=add_op.outputs[0], new_var=x
    )
    # Remove all the ops at once
    block.remove_ops([matmul_op, add_op])
    return True


def fuse_matmul_weight_bias_block(block):
    fusion_status = False
    for op in list(block.operations):
        for b in op.blocks:
            block_changed = True
            while block_changed:
                block_changed = fuse_matmul_weight_bias_block(b)
        if len(op.blocks) > 0:
            # This op can't be matmul
            continue

        add_op = match_pattern(op)
        if add_op is not None:
            with block:
                fusion_status = try_to_transform(op, add_op, block)
            # has to break as the downstream iterator is affected.
            if fusion_status:
                return fusion_status
    return fusion_status


@register_pass(namespace="common")
def fuse_matmul_weight_bias(prog):
    """
    Convert matmul + add/sub to linear whenever possible.

    Given:
        %3 = matmul(x=%1, y=%2)  # %1 or %2 is const and rank 2 (weight)
        ...
        %5 = add(x=%3, y=%4) # %4 is const. add(x=%4, y=%3) is equivalent
                             # sub is similar.

    Result:
        # assuming %2 above is const and rank 2
        %5 = linear(x=%1, weight=%2, bias=%4)

    Inputs:

        prog: Program
    """
    for f_name, f in prog.functions.items():
        block_changed = True
        while block_changed:
            block_changed = fuse_matmul_weight_bias_block(f)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import coremltools
from coremltools._deps import (
    _HAS_KERAS2_TF,
    MSG_KERAS2_NOT_FOUND,
    _HAS_SKLEARN,
    MSG_SKLEARN_NOT_FOUND,
)
from coremltools.models.utils import _macos_version, _is_macos

if _HAS_KERAS2_TF:
    import keras
if _HAS_SKLEARN:
    import sklearn
    from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
    from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
    from sklearn.linear_model import LinearRegression
    from sklearn.svm import SVC, SVR
    from sklearn.datasets import load_boston
import unittest
import numpy as np
import pytest
import PIL.Image


def create_model(spec):
    """
    Create MLModel with specified types
    Parameters
    ----------
    spec: Pb spec from 3rd party converted model

    Returns
    -------
    MLModel
    """
    return coremltools.models.MLModel(spec)


@unittest.skipUnless(
    _is_macos() and _macos_version() >= (10, 13), "Only supported on macOS 10.13+"
)
@unittest.skipIf(not _HAS_SKLEARN, MSG_SKLEARN_NOT_FOUND)
class TestIODataTypes(unittest.TestCase):
    """
    This class tests for different I/O feature data types for an .mlmodel
    It will cover the following areas to test for:
    - All features must have a valid type
    - Multiarrays must have a valid dataType. Inputs must specify shape. Shape must have >= 0 elements
    - Images must have a valid colorspace. width & height have to be >= 0
    - Dictionaries must have a valid key type
    """

    @property
    def scikit_data(self):
        return load_boston()

    def _feature_data_type(self, dtype):
        feature_dict = {np.int32: "INT32", np.float32: "FLOAT32", np.float64: "DOUBLE"}
        return feature_dict[dtype]

    @property
    def number_data_type(self):
        return dict(
            int8=np.int8,
            int16=np.int16,
            int32=np.int32,
            uint8=np.uint8,
            uint16=np.uint16,
            uint32=np.uint32,
            float=np.float32,
            double=np.double,
        )

    def _sklearn_setup(self, model, dtype, data, target):
        model.fit(data, target)
        spec = coremltools.converters.sklearn.convert(
            model, "data", "target"
        ).get_spec()
        return model, spec

    def _check_tree_model(self, spec, inputType, outputType, n_out):
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertIsNotNone(spec.treeEnsembleClassifier)

        # Test the interface class
        self.assertEqual(spec.description.predictedFeatureName, "target")

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.output), n_out)
        self.assertEqual(spec.description.output[0].name, "target")
        self.assertEqual(spec.description.output[0].type.WhichOneof("Type"), outputType)
        self.assertEqual(spec.description.input[0].name, "data")
        self.assertEqual(spec.description.input[0].type.WhichOneof("Type"), inputType)

    def test_tree_regressor(self):
        for dtype in self.number_data_type.keys():
            scikit_model = DecisionTreeRegressor(random_state=1)
            data = self.scikit_data["data"].astype(dtype)
            target = self.scikit_data["target"].astype(dtype)
            scikit_model, spec = self._sklearn_setup(scikit_model, dtype, data, target)
            test_data = data[0].reshape(1, -1)
            self._check_tree_model(spec, "multiArrayType", "doubleType", 1)
            coreml_model = create_model(spec)
            try:
                self.assertEqual(
                    scikit_model.predict(test_data)[0].dtype,
                    type(coreml_model.predict({"data": test_data})["target"]),
                )
                self.assertEqual(
                    scikit_model.predict(test_data)[0],
                    coreml_model.predict({"data": test_data})["target"],
                    msg="{} != {} for Dtype: {}".format(
                        scikit_model.predict(test_data)[0],
                        coreml_model.predict({"data": test_data})["target"],
                        dtype,
                    ),
                )
            except RuntimeError:
                print("{} not supported. ".format(dtype))

    def test_random_forest_classifier(self):
        for dtype in self.number_data_type.keys():
            scikit_model = RandomForestClassifier(random_state=1)
            data = self.scikit_data["data"].astype(dtype)
            target = (
                self.scikit_data["target"].astype(dtype)
                > self.scikit_data["target"].astype(dtype).mean()
            )
            scikit_model, spec = self._sklearn_setup(scikit_model, dtype, data, target)
            test_data = data[0].reshape(1, -1)
            self._check_tree_model(spec, "multiArrayType", "int64Type", 2)
            coreml_model = create_model(spec)
            try:
                self.assertEqual(
                    scikit_model.predict(test_data)[0],
                    bool(int(coreml_model.predict({"data": test_data})["target"])),
                    msg="{} != {} for Dtype: {}".format(
                        scikit_model.predict(test_data)[0],
                        bool(int(coreml_model.predict({"data": test_data})["target"])),
                        dtype,
                    ),
                )
            except RuntimeError:
                print("{} not supported. ".format(dtype))

    def test_random_forest_regressor(self):
        for dtype in self.number_data_type.keys():
            scikit_model = RandomForestRegressor(random_state=1)
            data = self.scikit_data["data"].astype(dtype)
            target = self.scikit_data["target"].astype(dtype)
            scikit_model, spec = self._sklearn_setup(scikit_model, dtype, data, target)
            test_data = data[0].reshape(1, -1)
            self._check_tree_model(spec, "multiArrayType", "doubleType", 1)
            coreml_model = create_model(spec)
            try:
                self.assertEqual(
                    scikit_model.predict(test_data)[0].dtype,
                    type(coreml_model.predict({"data": test_data})["target"]),
                )
                self.assertAlmostEqual(
                    scikit_model.predict(test_data)[0],
                    coreml_model.predict({"data": test_data})["target"],
                    msg="{} != {} for Dtype: {}".format(
                        scikit_model.predict(test_data)[0],
                        coreml_model.predict({"data": test_data})["target"],
                        dtype,
                    ),
                )
            except RuntimeError:
                print("{} not supported. ".format(dtype))

    def test_support_vector_classifier(self):
        for dtype in self.number_data_type.keys():
            scikit_model = SVC(kernel="rbf", gamma=1.2, C=1)
            data = self.scikit_data["data"].astype(dtype)
            target = (
                self.scikit_data["target"].astype(dtype)
                > self.scikit_data["target"].astype(dtype).mean()
            )
            scikit_model, spec = self._sklearn_setup(scikit_model, dtype, data, target)
            coreml_model = create_model(spec)
            for idx in range(0, 10):
                test_data = data[idx].reshape(1, -1)
                try:
                    self.assertEqual(
                        scikit_model.predict(test_data)[0],
                        bool(int(coreml_model.predict({"data": test_data})["target"])),
                        msg="{} != {} for Dtype: {}".format(
                            scikit_model.predict(test_data)[0],
                            bool(
                                int(coreml_model.predict({"data": test_data})["target"])
                            ),
                            dtype,
                        ),
                    )
                except RuntimeError:
                    print("{} not supported. ".format(dtype))

    def test_support_vector_regressor(self):
        for dtype in self.number_data_type.keys():
            scikit_model = SVR(kernel="rbf")
            data = self.scikit_data["data"].astype(dtype)
            target = self.scikit_data["target"].astype(dtype)
            scikit_model, spec = self._sklearn_setup(scikit_model, dtype, data, target)
            test_data = data[0].reshape(1, -1)
            coreml_model = create_model(spec)
            try:
                self.assertEqual(
                    scikit_model.predict(test_data)[0],
                    coreml_model.predict({"data": test_data})["target"],
                    msg="{} != {} for Dtype: {}".format(
                        scikit_model.predict(test_data)[0],
                        coreml_model.predict({"data": test_data})["target"],
                        dtype,
                    ),
                )
            except RuntimeError:
                print("{} not supported. ".format(dtype))

    def test_linear_regressor(self):
        for dtype in self.number_data_type.keys():
            scikit_model = LinearRegression(normalize=True)
            data = self.scikit_data["data"].astype(dtype)
            target = self.scikit_data["target"].astype(dtype)
            scikit_model, spec = self._sklearn_setup(scikit_model, dtype, data, target)
            test_data = data[0].reshape(1, -1)
            coreml_model = create_model(spec)
            try:
                self.assertEqual(
                    scikit_model.predict(test_data)[0].dtype,
                    type(coreml_model.predict({"data": test_data})["target"]),
                )
                self.assertAlmostEqual(
                    scikit_model.predict(test_data)[0],
                    coreml_model.predict({"data": test_data})["target"],
                    msg="{} != {} for Dtype: {}".format(
                        scikit_model.predict(test_data)[0],
                        coreml_model.predict({"data": test_data})["target"],
                        dtype,
                    ),
                )
            except RuntimeError:
                print("{} not supported. ".format(dtype))

    @unittest.skipIf(not _HAS_KERAS2_TF, MSG_KERAS2_NOT_FOUND)
    @pytest.mark.keras2
    def test_keras_dense_model(self):
        model = keras.models.Sequential()
        model.add(
            keras.layers.Dense(
                3,
                activation="sigmoid",
                kernel_initializer="random_uniform",
                bias_initializer="random_uniform",
                input_dim=3,
            )
        )
        for key, dtype in self.number_data_type.items():
            try:
                input_data = np.random.rand(1, 3).astype(key)
                keras_out = model.predict(input_data)
                coreml_model = coremltools.converters.keras.convert(
                    model, input_names=["data"], output_names=["target"]
                )
                spec = coreml_model.get_spec()
                spec.description.output[
                    0
                ].type.multiArrayType.dataType = coremltools.proto.FeatureTypes_pb2.ArrayFeatureType.ArrayDataType.Value(
                    self._feature_data_type(dtype)
                )
                spec.description.input[
                    0
                ].type.multiArrayType.dataType = coremltools.proto.FeatureTypes_pb2.ArrayFeatureType.ArrayDataType.Value(
                    self._feature_data_type(dtype)
                )
                coreml_model = coremltools.models.MLModel(spec)
                coreml_out = coreml_model.predict(
                    {"data": np.expand_dims(input_data, 0)}
                )["target"]
                self.assertEqual(dtype, coreml_out.dtype)
                if dtype != np.int32:
                    for idx in range(0, len(keras_out)):
                        self.assertAlmostEqual(
                            keras_out[0][idx],
                            coreml_out[idx],
                            msg="{}\n{} != {}".format(dtype, keras_out, coreml_out),
                            places=2,
                        )
            except KeyError:
                print("{} not supported. ".format(dtype))

    def test_keras_embedding_model(self):

        model = keras.models.Sequential()
        model.add(keras.layers.Embedding(100, 3, input_length=5, input_dtype="float32"))
        for key, dtype in self.number_data_type.items():
            try:
                input_data = np.random.randint(0, 100, size=(1, 5)).astype(key)
                keras_out = np.squeeze(model.predict(input_data)).flatten()
                coreml_model = coremltools.converters.keras.convert(
                    model, input_names=["data"], output_names=["target"]
                )

                spec = coreml_model.get_spec()
                spec.description.output[
                    0
                ].type.multiArrayType.dataType = coremltools.proto.FeatureTypes_pb2.ArrayFeatureType.ArrayDataType.Value(
                    self._feature_data_type(dtype)
                )
                spec.description.input[
                    0
                ].type.multiArrayType.dataType = coremltools.proto.FeatureTypes_pb2.ArrayFeatureType.ArrayDataType.Value(
                    self._feature_data_type(dtype)
                )
                coreml_model = coremltools.models.MLModel(spec)
                coreml_out = np.squeeze(
                    coreml_model.predict({"data": np.expand_dims(input_data, 0).T})[
                        "target"
                    ]
                ).flatten()
                self.assertEqual(dtype, coreml_out.dtype)
                if dtype != np.int32:
                    for idx in range(0, len(keras_out)):
                        self.assertAlmostEqual(
                            keras_out[idx],
                            coreml_out[idx],
                            msg="{}\n{} != {}".format(dtype, keras_out, coreml_out),
                            places=2,
                        )
            except KeyError:
                print("{} not supported. ".format(dtype))

    def test_image_output_rgb(self):
        input_shape = (3, 10, 20)
        input_features = [("data", coremltools.models.datatypes.Array(*input_shape))]
        output_features = [("target", coremltools.models.datatypes.Array(*input_shape))]
        builder = coremltools.models.neural_network.NeuralNetworkBuilder(
            input_features, output_features
        )
        builder.add_elementwise(
            "Identity",
            input_names=["data"],
            output_name="target",
            mode="ADD",
            alpha=0.0,
        )
        spec = builder.spec
        output = spec.description.output[0]
        output.type.imageType.colorSpace = coremltools.proto.FeatureTypes_pb2.ImageFeatureType.ColorSpace.Value(
            "RGB"
        )
        output.type.imageType.height = input_shape[1]
        output.type.imageType.width = input_shape[2]

        coreml_model = coremltools.models.MLModel(spec)
        input_data = np.floor(np.random.rand(*input_shape) * 255)

        coreml_out = coreml_model.predict({"data": input_data})["target"]
        self.assertEqual(PIL.Image.Image, type(coreml_out))
        self.assertEqual("RGBA", coreml_out.mode)
        np.testing.assert_equal(
            np.uint8(input_data), np.array(coreml_out).transpose(2, 0, 1)[:3, :]
        )

    @unittest.skip("rdar://71638164")
    def test_image_output_bgr(self):
        input_shape = (3, 15, 25)
        input_features = [("data", coremltools.models.datatypes.Array(*input_shape))]
        output_features = [("target", coremltools.models.datatypes.Array(*input_shape))]
        builder = coremltools.models.neural_network.NeuralNetworkBuilder(
            input_features, output_features
        )
        builder.add_elementwise(
            "Identity",
            input_names=["data"],
            output_name="target",
            mode="ADD",
            alpha=0.0,
        )
        spec = builder.spec
        output = spec.description.output[0]
        output.type.imageType.colorSpace = coremltools.proto.FeatureTypes_pb2.ImageFeatureType.ColorSpace.Value(
            "BGR"
        )
        output.type.imageType.height = input_shape[1]
        output.type.imageType.width = input_shape[2]

        coreml_model = coremltools.models.MLModel(spec)
        input_data = np.floor(np.random.rand(*input_shape) * 255)

        coreml_out = coreml_model.predict({"data": input_data})["target"]
        self.assertEqual(PIL.Image.Image, type(coreml_out))
        self.assertEqual("RGBA", coreml_out.mode)
        np.testing.assert_equal(
            np.uint8(input_data),
            np.array(coreml_out)[:, :, ::-1].transpose(2, 0, 1)[1:, :],
        )

    def test_image_output_grayscale(self):
        input_shape = (1, 20, 30)
        input_features = [("data", coremltools.models.datatypes.Array(*input_shape))]
        output_features = [("target", coremltools.models.datatypes.Array(*input_shape))]
        builder = coremltools.models.neural_network.NeuralNetworkBuilder(
            input_features, output_features
        )
        builder.add_elementwise(
            "Identity",
            input_names=["data"],
            output_name="target",
            mode="ADD",
            alpha=0.0,
        )
        spec = builder.spec
        output = spec.description.output[0]
        output.type.imageType.colorSpace = coremltools.proto.FeatureTypes_pb2.ImageFeatureType.ColorSpace.Value(
            "GRAYSCALE"
        )
        output.type.imageType.height = input_shape[1]
        output.type.imageType.width = input_shape[2]

        coreml_model = coremltools.models.MLModel(spec)
        input_data = np.floor(np.random.rand(*input_shape) * 255)

        coreml_out = coreml_model.predict({"data": input_data})["target"]
        self.assertEqual(PIL.Image.Image, type(coreml_out))
        self.assertEqual("L", coreml_out.mode)
        np.testing.assert_equal(np.uint8(input_data)[0], np.array(coreml_out))
